#! /usr/bin/env python
"""SearchSource: directory containing PDF files (based on GROBID)"""
from __future__ import annotations

import re
import typing
from pathlib import Path

import pymupdf
import requests
import zope.interface
from pydantic import Field

import colrev.env.local_index
import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.packages.pdf_backward_search.src.pdf_backward_search as bws
import colrev.record.qm.checkers.missing_field
import colrev.record.record
import colrev.record.record_pdf
import colrev.record.record_prep
import colrev.record.record_similarity
from colrev.constants import Colors
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import RecordState
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.crossref.src import crossref_api
from colrev.writer.write_utils import write_file

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class FilesSearchSource:
    """Files directories (PDFs based on GROBID)"""

    # pylint: disable=too-many-instance-attributes

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    endpoint = "colrev.files_dir"
    source_identifier = Fields.FILE
    search_types = [SearchType.FILES]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    _doi_regex = re.compile(r"10\.\d{4,9}/[-._;/:A-Za-z0-9]*")
    _batch_size = 20
    rerun: bool

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.review_manager = source_operation.review_manager
        self.source_operation = source_operation

        self.search_source = (
            colrev.package_manager.package_settings.DefaultSourceSettings(**settings)
        )

        if not self.review_manager.in_ci_environment():
            self.pdf_preparation_operation = self.review_manager.get_pdf_prep_operation(
                notify_state_transition_operation=False
            )

        self.pdfs_path = self.review_manager.path / Path(
            self.search_source.search_parameters["scope"]["path"]
        )

        self.subdir_pattern: re.Pattern = re.compile("")
        self.r_subdir_pattern: re.Pattern = re.compile("")
        if "subdir_pattern" in self.search_source.search_parameters.get("scope", {}):
            self.subdir_pattern = self.search_source.search_parameters["scope"][
                "subdir_pattern"
            ]
            self.review_manager.logger.info(
                f"Activate subdir_pattern: {self.subdir_pattern}"
            )
            if self.subdir_pattern == Fields.YEAR:
                self.r_subdir_pattern = re.compile("([1-3][0-9]{3})")
            if self.subdir_pattern == "volume_number":
                self.r_subdir_pattern = re.compile("([0-9]{1,3})(_|/)([0-9]{1,2})")
            if self.subdir_pattern == Fields.VOLUME:
                self.r_subdir_pattern = re.compile("([0-9]{1,4})")

        self.crossref_api = crossref_api.CrossrefAPI(params={})

    def _update_if_pdf_renamed(
        self,
        *,
        record_dict: dict,
        records: dict,
        search_source: Path,
    ) -> bool:
        updated = True
        not_updated = False

        c_rec_l = [
            r
            for r in records.values()
            if f"{search_source}/{record_dict['ID']}" in r[Fields.ORIGIN]
        ]
        if len(c_rec_l) == 1:
            c_rec = c_rec_l.pop()
            if "colrev_pdf_id" in c_rec:
                cpid = c_rec["colrev_pdf_id"]
                pdf_fp = self.review_manager.path / Path(record_dict[Fields.FILE])
                file_path = pdf_fp.parents[0]
                potential_pdfs = file_path.glob("*.pdf")

                for potential_pdf in potential_pdfs:
                    cpid_potential_pdf = colrev.record.record.Record.get_colrev_pdf_id(
                        potential_pdf,
                    )

                    if cpid == cpid_potential_pdf:
                        record_dict[Fields.FILE] = str(
                            potential_pdf.relative_to(self.review_manager.path)
                        )
                        c_rec[Fields.FILE] = str(
                            potential_pdf.relative_to(self.review_manager.path)
                        )
                        return updated
        return not_updated

    def _remove_records_if_pdf_no_longer_exists(self) -> None:
        # search_operation.review_manager.logger.debug(
        #     "Checking for PDFs that no longer exist"
        # )

        if not self.search_source.filename.is_file():
            return

        search_rd = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            logger=self.review_manager.logger,
            unique_id_field="ID",
        )

        records = self.review_manager.dataset.load_records_dict()

        to_remove: typing.List[str] = []
        files_removed = []
        for record_dict in search_rd.values():
            x_file_path = self.review_manager.path / Path(record_dict[Fields.FILE])
            if not x_file_path.is_file():
                if records:
                    updated = self._update_if_pdf_renamed(
                        record_dict=record_dict,
                        records=records,
                        search_source=self.search_source.filename,
                    )
                    if updated:
                        continue
                to_remove.append(
                    f"{self.search_source.filename.name}/{record_dict['ID']}"
                )
                files_removed.append(record_dict[Fields.FILE])

        search_rd = {
            x[Fields.ID]: x
            for x in search_rd.values()
            if (self.review_manager.path / Path(x[Fields.FILE])).is_file()
        }

        if len(search_rd.values()) != 0:

            write_file(records_dict=search_rd, filename=self.search_source.filename)

        if records:
            for record_dict in records.values():
                for origin_to_remove in to_remove:
                    if origin_to_remove in record_dict[Fields.ORIGIN]:
                        record_dict[Fields.ORIGIN].remove(origin_to_remove)
            if to_remove:
                self.review_manager.logger.info(
                    f" {Colors.RED}Removed {len(to_remove)} records "
                    f"(PDFs no longer available){Colors.END}"
                )
                print(" " + "\n ".join(files_removed))
            records = {k: v for k, v in records.items() if v[Fields.ORIGIN]}
            self.review_manager.dataset.save_records_dict(records)

    def _update_fields_based_on_pdf_dirs(
        self, *, record_dict: dict, params: dict
    ) -> dict:
        if not self.subdir_pattern:
            return record_dict

        if Fields.JOURNAL in params["scope"]:
            record_dict[Fields.JOURNAL] = params["scope"][Fields.JOURNAL]
            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE

        if "conference" in params["scope"]:
            record_dict[Fields.BOOKTITLE] = params["scope"]["conference"]
            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS

        if self.subdir_pattern:
            # Note : no file access here (just parsing the patterns)
            # no absolute paths needed
            partial_path = Path(record_dict[Fields.FILE]).parents[0]

            if self.subdir_pattern == Fields.YEAR:
                # Note: for year-patterns, we allow subfolders
                # (eg., conference tracks)
                match = self.r_subdir_pattern.search(str(partial_path))
                if match is not None:
                    year = match.group(1)
                    record_dict[Fields.YEAR] = year

            elif self.subdir_pattern == "volume_number":
                match = self.r_subdir_pattern.search(str(partial_path))
                if match is not None:
                    volume = match.group(1)
                    number = match.group(3)
                    record_dict[Fields.VOLUME] = volume
                    record_dict[Fields.NUMBER] = number
                else:
                    # sometimes, journals switch...
                    r_subdir_pattern = re.compile("([0-9]{1,3})")
                    match = r_subdir_pattern.search(str(partial_path))
                    if match is not None:
                        volume = match.group(1)
                        record_dict[Fields.VOLUME] = volume

            elif self.subdir_pattern == Fields.VOLUME:
                match = self.r_subdir_pattern.search(str(partial_path))
                if match is not None:
                    volume = match.group(1)
                    record_dict[Fields.VOLUME] = volume

        return record_dict

    def _get_missing_fields_from_doc_info(self, *, record_dict: dict) -> None:
        file_path = self.review_manager.path / Path(record_dict[Fields.FILE])
        doc = pymupdf.Document(file_path)

        # pylint: disable=no-member
        if record_dict.get(Fields.TITLE, "NA") in ["NA", ""]:
            if "title" in doc.metadata:
                try:
                    record_dict[Fields.TITLE] = doc.metadata["title"]
                except UnicodeDecodeError:
                    pass
        if record_dict.get(Fields.AUTHOR, "NA") in ["NA", ""]:
            if "author" in doc.metadata:
                try:
                    pdf_md_author = doc.metadata["author"]
                    if (
                        "Mirko Janc" not in pdf_md_author
                        and "wendy" != pdf_md_author
                        and "yolanda" != pdf_md_author
                    ):
                        record_dict[Fields.AUTHOR] = pdf_md_author
                except UnicodeDecodeError:
                    pass

    # curl -v --form input=@./profit.pdf localhost:8070/api/processHeaderDocument
    # curl -v --form input=@./thefile.pdf -H "Accept: application/x-bibtex"
    # -d "consolidateHeader=0" localhost:8070/api/processHeaderDocument
    def _get_record_from_pdf_grobid(self, *, record_dict: dict) -> dict:
        if RecordState.md_prepared == record_dict.get(Fields.STATUS, "NA"):
            return record_dict

        pdf_path = self.review_manager.path / Path(record_dict[Fields.FILE])
        try:
            tei = self.review_manager.get_tei(
                pdf_path=pdf_path,
            )
        except (
            FileNotFoundError,
            requests.exceptions.ReadTimeout,
            requests.exceptions.ConnectionError,
            colrev_exceptions.TEITimeoutException,
        ):
            return record_dict

        for key, val in tei.get_metadata().items():
            if val:
                record_dict[key] = str(val)

        self._get_missing_fields_from_doc_info(record_dict=record_dict)

        if Fields.ABSTRACT in record_dict:
            del record_dict[Fields.ABSTRACT]
        if Fields.KEYWORDS in record_dict:
            del record_dict[Fields.KEYWORDS]
        if Fields.DOI in record_dict:
            record_dict[Fields.DOI] = record_dict[Fields.DOI].upper()

        # to allow users to update/reindex with newer version:
        record_dict[Fields.GROBID_VERSION] = (
            "lfoppiano/grobid:" + tei.get_grobid_version()
        )

        return record_dict

    def _get_grobid_metadata(self, *, file_path: Path) -> dict:
        record_dict: typing.Dict[str, typing.Any] = {
            Fields.FILE: str(file_path),
            Fields.ENTRYTYPE: ENTRYTYPES.MISC,
        }
        try:
            record_dict = self._get_record_from_pdf_grobid(record_dict=record_dict)
            with pymupdf.open(file_path) as doc:
                pages_in_file = doc.page_count
                if pages_in_file < 6:
                    record = colrev.record.record_pdf.PDFRecord(
                        record_dict, path=self.review_manager.path
                    )
                    record.set_text_from_pdf()
                    record_dict = record.get_data()
                    if Fields.TEXT_FROM_PDF in record_dict:
                        text: str = record_dict[Fields.TEXT_FROM_PDF]
                        if "bookreview" in text.replace(" ", "").lower():
                            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC
                            record_dict["note"] = "Book review"
                        if "erratum" in text.replace(" ", "").lower():
                            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC
                            record_dict["note"] = "Erratum"
                        if "correction" in text.replace(" ", "").lower():
                            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC
                            record_dict["note"] = "Correction"
                        if "contents" in text.replace(" ", "").lower():
                            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC
                            record_dict["note"] = "Contents"
                        if "withdrawal" in text.replace(" ", "").lower():
                            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC
                            record_dict["note"] = "Withdrawal"
                        del record_dict[Fields.TEXT_FROM_PDF]
                    # else:
                    #     print(f'text extraction error in {record_dict[Fields.ID]}')

                record_dict = {k: v for k, v in record_dict.items() if v is not None}
                record_dict = {k: v for k, v in record_dict.items() if v != "NA"}

                # add details based on path
                record_dict = self._update_fields_based_on_pdf_dirs(
                    record_dict=record_dict, params=self.search_source.search_parameters
                )

        except colrev_exceptions.TEIException:
            pass

        return record_dict

    def _is_broken_filepath(
        self,
        file_path: Path,
    ) -> bool:
        if ";" in str(file_path):
            self.review_manager.logger.error(
                f'skipping PDF with ";" in filepath: \n{file_path}'
            )
            return True

        if (
            "_ocr.pdf" == str(file_path)[-8:]
            or "_with_cp.pdf" == str(file_path)[-10:]
            or "_with_lp.pdf" == str(file_path)[-10:]
            or "_backup.pdf" == str(file_path)[-11:]
        ):
            self.review_manager.logger.info(
                f"Skipping PDF with _ocr.pdf/_with_cp.pdf: {file_path}"
            )
            return True

        return False

    def _validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""

        source = self.search_source

        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        assert source.search_type == SearchType.FILES

        if "subdir_pattern" in source.search_parameters:
            if source.search_parameters["subdir_pattern"] != [
                "NA",
                "volume_number",
                Fields.YEAR,
                Fields.VOLUME,
            ]:
                raise colrev_exceptions.InvalidQueryException(
                    "subdir_pattern not in [NA, volume_number, year, volume]"
                )

        if "sub_dir_pattern" in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "sub_dir_pattern: deprecated. use subdir_pattern"
            )

        if "scope" not in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "scope required in search_parameters"
            )
        if "path" not in source.search_parameters["scope"]:
            raise colrev_exceptions.InvalidQueryException(
                "path required in search_parameters/scope"
            )
        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def _add_md_string(self, *, record_dict: dict) -> dict:
        # To identify potential duplicates
        if Path(record_dict[Fields.FILE]).suffix != ".pdf":
            return record_dict

        md_copy = record_dict.copy()
        try:
            fsize = str(
                (self.review_manager.path / Path(record_dict[Fields.FILE]))
                .stat()
                .st_size
            )
        except FileNotFoundError:
            fsize = "NOT_FOUND"
        for key in [Fields.ID, Fields.GROBID_VERSION, Fields.FILE]:
            if key in md_copy:
                md_copy.pop(key)
        md_string = ",".join([f"{k}:{v}" for k, v in md_copy.items()])
        record_dict["md_string"] = str(fsize) + md_string
        return record_dict

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _index_file(
        self,
        *,
        file_path: Path,
        files_dir_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        linked_file_paths: list,
        local_index: colrev.env.local_index.LocalIndex,
    ) -> dict:
        if file_path.suffix == ".pdf":
            return self._index_pdf(
                file_path=file_path,
                files_dir_feed=files_dir_feed,
                linked_file_paths=linked_file_paths,
                local_index=local_index,
            )
        if file_path.suffix == ".mp4":
            return self._index_mp4(
                file_path=file_path,
                files_dir_feed=files_dir_feed,
                linked_file_paths=linked_file_paths,
                local_index=local_index,
            )
        raise NotImplementedError

    def _fix_grobid_errors(self, new_record: dict) -> None:
        # Fix common GROBID errors that would cause problems in deduplication

        # drop title if it is identical with journal
        if Fields.TITLE in new_record and Fields.JOURNAL in new_record:
            if new_record[Fields.TITLE] == new_record[Fields.JOURNAL]:
                new_record.pop(Fields.TITLE)
        # drop title if it starts with "doi:"
        if Fields.TITLE in new_record:
            if new_record[Fields.TITLE].lower().startswith("doi:"):
                new_record.pop(Fields.TITLE)

        # drop title with erroneous terms
        if Fields.TITLE in new_record:
            if any(
                x in new_record[Fields.TITLE].lower()
                for x in [
                    "papers must be in english",
                    "please contact",
                    "before submission",
                    "chairman of the editorial board",
                    "received his ph",
                    "received her ph",
                    "can be submitted to",
                ]
            ):
                new_record.pop(Fields.TITLE)
        # drop title based on erroneous list
        if Fields.TITLE in new_record:
            if new_record[Fields.TITLE].lower() in [
                "the international journal of information systems applications",
                "c ommunications of the a i s ssociation for nformation ystems",
                "communications of the association for information systems "
                + "communications of the association for information systems",
            ]:
                new_record.pop(Fields.TITLE)

        # drop title if longer than 200 characters
        if Fields.TITLE in new_record:
            if len(new_record[Fields.TITLE]) > 200:
                new_record.pop(Fields.TITLE)

        # drop title if it has more numbers than characters
        if Fields.TITLE in new_record:
            if sum(c.isdigit() for c in new_record[Fields.TITLE]) > sum(
                c.isalpha() for c in new_record[Fields.TITLE]
            ):
                new_record.pop(Fields.TITLE)

    def _index_pdf(
        self,
        *,
        file_path: Path,
        files_dir_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        linked_file_paths: list,
        local_index: colrev.env.local_index.LocalIndex,
    ) -> dict:
        new_record: dict = {}

        file_path_abs = self.review_manager.path / file_path
        if self._is_broken_filepath(file_path=file_path_abs):
            return new_record

        if not self.review_manager.force_mode:
            # note: for curations, we want all pdfs indexed/merged separately,
            # in other projects, it is generally sufficient if the pdf is linked
            if not self.review_manager.settings.is_curated_masterdata_repo():
                if file_path in linked_file_paths:
                    # Otherwise: skip linked PDFs
                    return new_record

        if not self.rerun:

            if file_path in [
                Path(r[Fields.FILE])
                for r in files_dir_feed.feed_records.values()
                if Fields.FILE in r
            ]:
                return new_record
        # otherwise: reindex all

        self.review_manager.logger.info(f" extract metadata from {file_path}")
        try:
            if not self.review_manager.settings.is_curated_masterdata_repo():
                # retrieve_based_on_colrev_pdf_id
                colrev_pdf_id = colrev.record.record.Record.get_colrev_pdf_id(
                    pdf_path=file_path_abs
                )
                new_record_object = local_index.retrieve_based_on_colrev_pdf_id(
                    colrev_pdf_id=colrev_pdf_id
                )
                new_record = new_record_object.data
                # Note : an alternative to replacing all data with the curated version
                # is to just add the curation_ID
                # (and retrieve the curated metadata separately/non-redundantly)
            else:
                new_record = self._get_grobid_metadata(file_path=file_path)
        except FileNotFoundError:
            self.review_manager.logger.error(f"File not found: {file_path} (skipping)")
            return {}
        except (
            colrev_exceptions.PDFHashError,
            colrev_exceptions.RecordNotInIndexException,
        ):
            # otherwise, get metadata from grobid (indexing)
            new_record = self._get_grobid_metadata(file_path=file_path_abs)

        self._fix_grobid_errors(new_record)

        new_record[Fields.FILE] = str(file_path)
        new_record = self._add_md_string(record_dict=new_record)

        # Note: identical md_string as a heuristic for duplicates
        potential_duplicates = [
            r
            for r in files_dir_feed.feed_records.values()
            if r["md_string"] == new_record["md_string"]
            and not r[Fields.FILE] == new_record[Fields.FILE]
        ]
        if potential_duplicates:
            self.review_manager.logger.warning(
                f" {Colors.RED}skip record (PDF potential duplicate): "
                f"{new_record['file']} {Colors.END} "
                f"({','.join([r['file'] for r in potential_duplicates])})"
            )

        return new_record

    def _index_mp4(
        self,
        *,
        file_path: Path,
        files_dir_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        linked_file_paths: list,
        local_index: colrev.env.local_index.LocalIndex,
    ) -> dict:
        record_dict = {Fields.ENTRYTYPE: "online", Fields.FILE: file_path}
        return record_dict

    def _get_file_batches(self) -> list:
        types = ("**/*.pdf", "**/*.mp4")
        files_grabbed: typing.List[Path] = []
        for suffix in types:
            files_grabbed.extend(self.pdfs_path.glob(suffix))

        files_to_index = [
            x.relative_to(self.review_manager.path) for x in files_grabbed
        ]

        file_batches = [
            files_to_index[i * self._batch_size : (i + 1) * self._batch_size]
            for i in range(
                (len(files_to_index) + self._batch_size - 1) // self._batch_size
            )
        ]
        return file_batches

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-nested-blocks
    # pylint: disable=too-many-locals
    def _run_dir_search(
        self,
        *,
        files_dir_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        local_index: colrev.env.local_index.LocalIndex,
        linked_file_paths: list,
    ) -> None:
        file_batches = self._get_file_batches()
        if not file_batches:
            files_dir_feed.save()
            return
        for i, file_batch in enumerate(file_batches):
            for record in files_dir_feed.feed_records.values():
                record = self._add_md_string(record_dict=record)

            for file_path in file_batch:
                new_record = self._index_file(
                    file_path=file_path,
                    files_dir_feed=files_dir_feed,
                    linked_file_paths=linked_file_paths,
                    local_index=local_index,
                )
                if new_record == {}:
                    continue

                self._add_doi_from_pdf_if_not_available(new_record)
                retrieved_record = colrev.record.record.Record(new_record)

                # Generally: add to feed but do not "update" records
                prev_feed_record = files_dir_feed.get_prev_feed_record(retrieved_record)
                files_dir_feed.add_record_to_feed(retrieved_record, prev_feed_record)

                if self.rerun:
                    # If rerun: fix_grobid_fields: in feed and records (if only pdf-file origin)
                    prefix = self.search_source.get_origin_prefix()
                    origin = f"{prefix}/{retrieved_record.data['ID']}"
                    for record_dict in files_dir_feed.records.values():
                        if origin in record_dict[Fields.ORIGIN]:
                            if len(record_dict[Fields.ORIGIN]) == 1:
                                self._fix_grobid_errors(record_dict)
                                if Fields.TITLE not in record_dict:
                                    record_dict[Fields.STATUS] = (
                                        RecordState.md_needs_manual_preparation
                                    )

            for record in files_dir_feed.feed_records.values():
                record.pop("md_string")

            last_round = i == len(file_batches) - 1
            files_dir_feed.save(skip_print=not last_round)

    def _add_doi_from_pdf_if_not_available(self, record_dict: dict) -> None:
        if Path(record_dict[Fields.FILE]).suffix != ".pdf":
            return
        try:
            record = colrev.record.record_pdf.PDFRecord(
                record_dict, path=self.review_manager.path
            )
            if Fields.DOI not in record_dict:
                record.set_text_from_pdf()
                res = re.findall(self._doi_regex, record.data[Fields.TEXT_FROM_PDF])
                if res:
                    record.data[Fields.DOI] = res[0].upper()
            record.data.pop(Fields.TEXT_FROM_PDF, None)
            record.data.pop(Fields.NR_PAGES_IN_FILE, None)
        except colrev_exceptions.InvalidPDFException:
            pass

    def search(self, rerun: bool) -> None:
        """Run a search of a Files directory"""

        self.rerun = rerun
        self._validate_source()

        # Do not run in continuous-integration environment
        if self.review_manager.in_ci_environment():
            raise colrev_exceptions.SearchNotAutomated("PDFs Dir Search not automated.")

        if self.review_manager.force_mode:  # i.e., reindex all
            self.review_manager.logger.info("Reindex all")

        # Removing records/origins for which PDFs were removed makes sense for curated repositories
        # In regular repositories, it may be confusing (e.g., if PDFs are renamed)
        # In these cases, we may simply print a warning instead of modifying/removing records?
        if self.review_manager.settings.is_curated_masterdata_repo():
            self._remove_records_if_pdf_no_longer_exists()

        grobid_service = self.review_manager.get_grobid_service()
        grobid_service.start()

        local_index = colrev.env.local_index.LocalIndex()

        records = self.review_manager.dataset.load_records_dict()
        files_dir_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        linked_file_paths = [
            Path(r[Fields.FILE]) for r in records.values() if Fields.FILE in r
        ]

        self._run_dir_search(
            files_dir_feed=files_dir_feed,
            linked_file_paths=linked_file_paths,
            local_index=local_index,
        )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for PDF directories (GROBID)"""

        result = {"confidence": 0.0}

        if filename.suffix == ".pdf" and not bws.BackwardSearchSource.heuristic(
            filename=filename, data=data
        ):
            result["confidence"] = 1.0
            return result

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        filename = operation.get_unique_filename(file_path_string="files")
        # pylint: disable=no-value-for-parameter
        search_source = colrev.settings.SearchSource(
            endpoint="colrev.files_dir",
            filename=filename,
            search_type=SearchType.FILES,
            search_parameters={"scope": {"path": "data/pdfs"}},
            comment="",
        )
        operation.add_source_and_search(search_source)
        return search_source

    def _update_based_on_doi(self, *, record_dict: dict) -> None:
        if Fields.DOI not in record_dict:
            return
        try:
            retrieved_record = self.crossref_api.query_doi(
                doi=record_dict[Fields.DOI],
            )

            if not colrev.record.record_similarity.matches(
                colrev.record.record.Record(record_dict), retrieved_record
            ):
                del record_dict[Fields.DOI]
                return

            for key in [
                Fields.JOURNAL,
                Fields.BOOKTITLE,
                Fields.VOLUME,
                Fields.NUMBER,
                Fields.YEAR,
                Fields.PAGES,
            ]:
                if key in retrieved_record.data:
                    record_dict[key] = retrieved_record.data[key]
        except (colrev_exceptions.RecordNotFoundInPrepSourceException,):
            pass

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":

            def field_mapper(record_dict: dict) -> None:
                if "note" in record_dict:
                    record_dict[f"{self.endpoint}.note"] = record_dict.pop("note")

            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                unique_id_field="ID",
                field_mapper=field_mapper,
                logger=self.review_manager.logger,
            )

            for record_dict in records.values():
                if Fields.GROBID_VERSION in record_dict:
                    del record_dict[Fields.GROBID_VERSION]

                # self._update_based_on_doi(record_dict=record_dict)

                # Rerun restrictions and __update_fields_based_on_pdf_dirs
                # because the restrictions/subdir-pattern may change
                record_dict = self._update_fields_based_on_pdf_dirs(
                    record_dict=record_dict, params=self.search_source.search_parameters
                )

            return records

        raise NotImplementedError

    def _fix_special_chars(self, *, record: colrev.record.record.Record) -> None:
        # We may also apply the following upon loading tei content
        if Fields.TITLE in record.data:
            record.data[Fields.TITLE] = (
                record.data[Fields.TITLE]
                .replace("n ˜", "ñ")
                .replace("u ´", "ú")
                .replace("ı ´", "í")
                .replace("a ´", "á")
                .replace("o ´", "ó")
                .replace("e ´", "é")
                .replace("c ¸", "ç")
                .replace("a ˜", "ã")
            )

        if Fields.AUTHOR in record.data:
            record.data[Fields.AUTHOR] = (
                record.data[Fields.AUTHOR]
                .replace("n ˜", "ñ")
                .replace("u ´", "ú")
                .replace("ı ´", "í")
                .replace("a ´", "á")
                .replace("o ´", "ó")
                .replace("e ´", "é")
                .replace("c ¸", "ç")
                .replace("a ˜", "ã")
            )

    def _fix_title_suffix(self, *, record: colrev.record.record.Record) -> None:
        if Fields.TITLE not in record.data:
            return
        if record.data[Fields.TITLE].endswith("Formula 1"):
            return
        if re.match(r"\d{4}$", record.data[Fields.TITLE]):
            return
        if record.data.get(Fields.TITLE, "").endswith(" 1"):
            record.data[Fields.TITLE] = record.data[Fields.TITLE][:-2]

    def _fix_special_outlets(self, *, record: colrev.record.record.Record) -> None:
        # Erroneous suffixes in IS conferences
        if record.data.get(Fields.BOOKTITLE, "") in [
            "Americas Conference on Information Systems",
            "International Conference on Information Systems",
            "European Conference on Information Systems",
            "Pacific Asia Conference on Information Systems",
        ]:
            for suffix in [
                "completed research paper",
                "completed research",
                "complete research",
                "full research paper",
                "research in progress",
                "(research in progress)",
            ]:
                if record.data[Fields.TITLE].lower().endswith(suffix):
                    record.data[Fields.TITLE] = record.data[Fields.TITLE][
                        : -len(suffix)
                    ].rstrip(" -:")
        # elif ...

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
        source: colrev.settings.SearchSource,
    ) -> colrev.record.record.Record:
        """Source-specific preparation for files"""

        if Fields.FILE not in record.data:
            return record

        if Path(record.data[Fields.FILE]).suffix == ".pdf":
            record.format_if_mostly_upper(Fields.TITLE, case="sentence")
            record.format_if_mostly_upper(Fields.JOURNAL, case=Fields.TITLE)
            record.format_if_mostly_upper(Fields.BOOKTITLE, case=Fields.TITLE)
            record.format_if_mostly_upper(Fields.AUTHOR, case=Fields.TITLE)

            if Fields.AUTHOR in record.data:
                record.data[Fields.AUTHOR] = (
                    record.data[Fields.AUTHOR]
                    .replace(" and T I C L E I N F O, A. R", "")
                    .replace(" and Quarterly, Mis", "")
                )

            # Typical error in old papers: title fields are equal to journal/booktitle fields
            if (
                record.data.get(Fields.TITLE, "no_title").lower()
                == record.data.get(Fields.JOURNAL, "no_journal").lower()
            ):
                record.remove_field(key=Fields.TITLE, source="files_dir_prepare")
                record.set_status(RecordState.md_needs_manual_preparation)
            if record.data.get(Fields.TITLE, "no_title").lower() == record.data.get(
                Fields.BOOKTITLE, "no_booktitle"
            ):
                record.remove_field(key=Fields.TITLE, source="files_dir_prepare")
                record.set_status(RecordState.md_needs_manual_preparation)
            self._fix_title_suffix(record=record)
            self._fix_special_chars(record=record)
            self._fix_special_outlets(record=record)

        return record



#! /usr/bin/env python
"""CoLRev pdf_get_man operation: Get PDF documents manually."""
from __future__ import annotations

import csv
import typing
from pathlib import Path

import pandas as pd

import colrev.exceptions as colrev_exceptions
import colrev.process.operation
import colrev.record.record
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.constants import RecordState


class PDFGetMan(colrev.process.operation.Operation):
    """Get PDFs manually"""

    pdf_get_man_package_endpoints: dict[str, typing.Any]
    MISSING_PDF_FILES_RELATIVE = Path("data/pdf_get_man/missing_pdf_files.csv")

    type = OperationsType.pdf_get_man

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.missing_pdf_files_csv = (
            self.review_manager.path / self.MISSING_PDF_FILES_RELATIVE
        )
        self.verbose = True

    def get_pdf_get_man(self, records: dict) -> list:
        """Get the records that are missing a PDF"""
        missing_records = []
        for record in records.values():
            if record[Fields.STATUS] == RecordState.pdf_needs_manual_retrieval:
                missing_records.append(record)
        return missing_records

    def export_retrieval_table(self, records: dict) -> None:
        """Export a table for manual PDF retrieval"""

        missing_records = self.get_pdf_get_man(records)

        if len(missing_records) > 0:
            missing_records_df = pd.DataFrame.from_records(missing_records)
            # pylint: disable=duplicate-code
            col_order = [
                Fields.ID,
                Fields.AUTHOR,
                Fields.TITLE,
                Fields.JOURNAL,
                Fields.BOOKTITLE,
                Fields.YEAR,
                Fields.VOLUME,
                Fields.NUMBER,
                Fields.PAGES,
                Fields.DOI,
            ]
            missing_records_df = missing_records_df.reindex(col_order, axis=1)
            self.missing_pdf_files_csv.parent.mkdir(exist_ok=True, parents=True)
            missing_records_df.to_csv(
                self.missing_pdf_files_csv, index=False, quoting=csv.QUOTE_ALL
            )

            self.review_manager.logger.info(
                f"Created {self.missing_pdf_files_csv} with paper details"
            )

    def discard(self) -> None:
        """Discard missing PDFs (set to pdf_not_available)"""

        records = self.review_manager.dataset.load_records_dict()
        for record_dict in records.values():
            record = colrev.record.record.Record(record_dict)
            if record.data[Fields.STATUS] == RecordState.pdf_needs_manual_retrieval:
                record.set_status(RecordState.pdf_not_available)
        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(
            msg="Discard missing PDFs", manual_author=True
        )

    def get_data(self) -> dict:
        """Get the data for pdf-get-man"""
        # pylint: disable=duplicate-code

        pdf_dir = self.review_manager.paths.pdf
        pdf_dir.mkdir(exist_ok=True)

        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())
        nr_tasks = len(
            [
                x
                for x in record_header_list
                if RecordState.pdf_needs_manual_retrieval == x[Fields.STATUS]
            ]
        )
        pad = min((max(len(x[Fields.ID]) for x in record_header_list) + 2), 40)
        items = self.review_manager.dataset.read_next_record(
            conditions=[{Fields.STATUS: RecordState.pdf_needs_manual_retrieval}]
        )
        pdf_get_man_data = {"nr_tasks": nr_tasks, "PAD": pad, "items": items}
        self.review_manager.logger.debug(
            self.review_manager.p_printer.pformat(pdf_get_man_data)
        )
        return pdf_get_man_data

    def pdfs_retrieved_manually(self) -> bool:
        """Check whether PDFs were retrieved manually"""
        return self.review_manager.dataset.has_record_changes()

    def pdf_get_man_record(
        self,
        *,
        record: colrev.record.record.Record,
        filepath: typing.Optional[Path] = None,
        PAD: int = 40,
    ) -> None:
        """Record pdf-get-man decision"""
        if filepath is not None:
            record.set_status(RecordState.pdf_imported)
            record.data.update(file=str(filepath.relative_to(self.review_manager.path)))
            self.review_manager.report_logger.info(
                f" {record.data['ID']}".ljust(PAD, " ") + "retrieved and linked PDF"
            )
            self.review_manager.logger.info(
                f" {record.data['ID']}".ljust(PAD, " ") + "retrieved and linked PDF"
            )
        else:
            if (
                self.review_manager.settings.pdf_get.pdf_required_for_screen_and_synthesis
            ):
                record.set_status(RecordState.pdf_not_available)
                self.review_manager.report_logger.info(
                    f" {record.data['ID']}".ljust(PAD, " ")
                    + "recorded as not_available"
                )
                self.review_manager.logger.info(
                    f" {record.data['ID']}".ljust(PAD, " ")
                    + "recorded as not_available"
                )
            else:
                record.set_status(RecordState.pdf_prepared)

                record.add_field_provenance(
                    key=Fields.FILE, source="pdf-get-man", note="not_available"
                )

                self.review_manager.report_logger.info(
                    f" {record.data['ID']}".ljust(PAD, " ")
                    + "recorded as not_available (and moved to screen)"
                )
                self.review_manager.logger.info(
                    f" {record.data['ID']}".ljust(PAD, " ")
                    + "recorded as not_available (and moved to screen)"
                )

        record_dict = record.get_data()
        self.review_manager.dataset.save_records_dict(
            {record_dict[Fields.ID]: record_dict}, partial=True
        )

    @colrev.process.operation.Operation.decorate()
    def main(self) -> None:
        """Get PDFs manually (main entrypoint)"""

        if (
            self.review_manager.in_ci_environment()
            and not self.review_manager.in_test_environment()
        ):
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="colrev pdf-get-man",
                detailed_trace="pdf-get-man not available in ci environment",
            )

        records = self.review_manager.dataset.load_records_dict()
        pdf_get_man_package_endpoints = (
            self.review_manager.settings.pdf_get.pdf_get_man_package_endpoints
        )

        package_manager = self.review_manager.get_package_manager()

        for pdf_get_man_package_endpoint in pdf_get_man_package_endpoints:
            pdf_get_man_class = package_manager.get_package_endpoint_class(
                package_type=EndpointType.pdf_get_man,
                package_identifier=pdf_get_man_package_endpoint["endpoint"],
            )
            endpoint = pdf_get_man_class(
                pdf_get_man_operation=self, settings=pdf_get_man_package_endpoint
            )

            records = endpoint.pdf_get_man(records)  # type: ignore



#! /usr/bin/env python
"""Checker for name particles."""
from __future__ import annotations

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class NameParticlesChecker:
    """The NameParticlesChecker"""

    fields_to_check = [Fields.AUTHOR, Fields.EDITOR]

    msg = DefectCodes.NAME_PARTICLES

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the name-particles checks"""

        for key in self.fields_to_check:
            if key not in record.data or record.ignored_defect(
                key=key, defect=self.msg
            ):
                continue
            if record.masterdata_is_curated():
                continue

            names = record.data[key].split(" and ")
            if any(self._particle_defect(name=name) for name in names):
                record.add_field_provenance_note(key=key, note=self.msg)
            else:
                record.remove_field_provenance_note(key=key, note=self.msg)

    def _particle_defect(self, *, name: str) -> bool:
        if name.endswith(" vom") or name.startswith("vom "):
            return True

        if name.endswith(" von") or name.startswith("von "):
            return True

        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(NameParticlesChecker(quality_model))



#! /usr/bin/env python
"""Discovering and using packages."""
from __future__ import annotations

import json
import os
import subprocess
import tempfile
from pathlib import Path

import pkg_resources
import toml


# pylint: disable=too-many-return-statements
def _get_local_editable_colrev_path() -> str:

    try:
        dist = pkg_resources.get_distribution("colrev")
    except pkg_resources.DistributionNotFound:
        print("CoLRev not installed")
        return ""

    # Construct the .dist-info directory name
    dist_info_folder = f"{dist.project_name.replace('-', '_')}"
    dist_info_folder += f"-{dist.version}.dist-info"
    if not dist.location:
        raise ValueError
    dist_info_folder = os.path.join(dist.location, dist_info_folder)

    direct_url_path = os.path.join(dist_info_folder, "direct_url.json")
    if not os.path.exists(direct_url_path):
        return ""

    with open(direct_url_path, encoding="utf-8") as file:
        data = json.load(file)

    if "url" not in data:
        return ""

    if not data["url"].startswith("file://"):
        return ""

    editable_dir = data["url"].replace("file://", "")
    return editable_dir


def _clone_colrev_repository() -> Path:
    temp_dir = tempfile.mkdtemp()
    print(f"Clone to {temp_dir}")
    subprocess.run(
        [
            "git",
            "clone",
            "--depth",
            "1",
            "https://github.com/CoLRev-Environment/colrev",
            temp_dir,
        ],
        check=False,
    )
    return Path(temp_dir)


def _get_colrev_path() -> Path:

    local_editable_colrev_path = _get_local_editable_colrev_path()
    if local_editable_colrev_path:
        colrev_path = Path(local_editable_colrev_path)
        if local_editable_colrev_path.startswith("/D:"):
            # Remove the leading slash to correct the path format
            corrected_path_str = local_editable_colrev_path.lstrip("/")
            colrev_path = Path(corrected_path_str).resolve(strict=False)
        else:
            colrev_path = colrev_path.resolve(strict=False)
    else:
        colrev_path = _clone_colrev_repository().resolve(strict=False)

    # Check if there is a nested colrev directory
    potential_nested_path = colrev_path / "colrev"
    if potential_nested_path.is_dir():
        colrev_path = potential_nested_path

    return colrev_path


def get_internal_packages_dict() -> dict:
    """Get a dictionary of internal CoLRev packages"""

    colrev_path = _get_colrev_path()
    packages_dir = colrev_path / Path("packages")

    internal_packages_dict = {}
    for package_dir in packages_dir.iterdir():
        if package_dir.is_dir():
            package_path = str(package_dir)

            pyproject_path = os.path.join(package_path, "pyproject.toml")
            if not os.path.exists(pyproject_path):
                continue
            with open(pyproject_path, encoding="utf-8") as file:
                pyproject_data = toml.load(file)
            package_name = pyproject_data["tool"]["poetry"]["name"]

            internal_packages_dict[package_name] = package_path

    return internal_packages_dict



#! /usr/bin/env python
"""Template for a custom Screen PackageEndpoint"""
from __future__ import annotations

import random
import typing

import zope.interface

import colrev.package_manager.interfaces
import colrev.package_manager.package_settings
import colrev.process.operation
import colrev.record.record
from colrev.constants import Fields

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.screen.Screen

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.ScreenInterface)
class CustomScreen:
    """Class for custom screen scripts"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    def __init__(
        self,
        *,
        screen_operation: colrev.screen.Screen,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)

    def run_screen(
        self, screen_operation: colrev.screen.Screen, records: dict, split: list
    ) -> dict:
        """Screen a record"""

        screen_data = screen_operation.get_data()
        screening_criteria = screen_operation.review_manager.settings.screen.criteria

        screening_criteria_available = bool(screening_criteria)

        for record_dict in screen_data["items"]:
            if len(split) > 0:
                if record_dict[Fields.ID] not in split:
                    continue

            record = colrev.record.record.Record(record_dict)

            if random.random() < 0.5:  # nosec
                if screening_criteria_available:
                    # record criteria
                    pass
                screen_operation.screen(
                    record=record,
                    screen_inclusion=True,
                    screening_criteria="...",
                )

            else:
                if screening_criteria_available:
                    # record criteria
                    pass
                screen_operation.screen(
                    record=record,
                    screen_inclusion=False,
                    screening_criteria="...",
                )

        screen_operation.review_manager.dataset.save_records_dict(records)
        screen_operation.review_manager.dataset.add_record_changes()
        screen_operation.review_manager.dataset.create_commit(
            msg="Screen (random)", manual_author=False, script_call="colrev screen"
        )
        return records



#! /usr/bin/env python
"""Service parsing metadata from PDFs/TEIs (created by GROBID)."""
from __future__ import annotations

import re
import typing
from pathlib import Path

import requests
from lxml import etree
from lxml.etree import XMLSyntaxError  # nosec

import colrev.env.grobid_service
import colrev.exceptions as colrev_exceptions
import colrev.process.operation
import colrev.record.record
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import RecordState

# xpath alternative:
# tree.xpath("//tei:sourceDesc/tei:biblStruct/tei:monogr/tei:idno",
# namespaces={"tei": "http://www.tei-c.org/ns/1.0"})
# abstract_node =tree.xpath("//tei:profileDesc/tei:abstract",
# namespaces={"tei": "http://www.tei-c.org/ns/1.0"})
# etree.tostring(abstract_node[0]).decode("utf-8")

if typing.TYPE_CHECKING:  # pragma: no cover
    from xml.etree.ElementTree import Element  # nosec


class TEIParser:
    """Environment service for TEI parsing"""

    ns = {
        "tei": "{http://www.tei-c.org/ns/1.0}",
        "w3": "{http://www.w3.org/XML/1998/namespace}",
    }
    nsmap = {
        "tei": "http://www.tei-c.org/ns/1.0",
        "w3": "http://www.w3.org/XML/1998/namespace",
    }

    def __init__(
        self,
        *,
        environment_manager: colrev.env.environment_manager.EnvironmentManager,
        pdf_path: typing.Optional[Path] = None,
        tei_path: typing.Optional[Path] = None,
    ):
        """Creates a TEI file
        modes of operation:
        - pdf_path: create TEI and temporarily store in self.data
        - pfd_path and tei_path: create TEI and save in tei_path
        - tei_path: read TEI from file
        """

        self.environment_manager = environment_manager
        # pylint: disable=consider-using-with
        assert pdf_path is not None or tei_path is not None
        if pdf_path is not None:
            if pdf_path.is_symlink():  # pragma: no cover
                pdf_path = pdf_path.resolve()
        self.pdf_path = pdf_path
        self.tei_path = tei_path
        if pdf_path is not None and not pdf_path.is_file():
            raise FileNotFoundError

        load_from_tei = False
        if tei_path is not None:
            if tei_path.is_file():
                load_from_tei = True

        if pdf_path is not None and not load_from_tei:
            self._create_tei()

        elif tei_path is not None:
            self.root = self._read_from_tei()  # type: ignore

    def _read_from_tei(self):  # type: ignore
        """Read a TEI from file"""
        with open(self.tei_path, "rb") as data:
            xslt_content = data.read()

        if b"[BAD_INPUT_DATA]" in xslt_content[:100]:
            raise colrev_exceptions.TEIException()

        return etree.XML(xslt_content)

    def _create_tei(self) -> None:
        """Create the TEI (based on GROBID)"""
        grobid_service = colrev.env.grobid_service.GrobidService(
            environment_manager=self.environment_manager
        )
        grobid_service.start()
        # Note: we have more control and transparency over the consolidation
        # if we do it in the colrev process
        options = {"consolidateHeader": "0", "consolidateCitations": "0"}

        # Note: Grobid offers direct export of Bibtex:
        # r = requests.post(
        #     GROBID_SERVICE.GROBID_URL() + "/api/processHeaderDocument",
        #     headers={"Accept": "application/x-bibtex"},
        # But parsing the metadata from the tei gives us more control of the details

        try:
            # pylint: disable=consider-using-with
            ret = requests.post(
                grobid_service.GROBID_URL + "/api/processFulltextDocument",
                files={"input": open(str(self.pdf_path), "rb")},
                data=options,
                timeout=180,
            )

            # Possible extension: get header only (should be more efficient)
            # r = requests.post(
            #     GrobidService.GROBID_URL + "/api/processHeaderDocument",
            #     files=dict(input=open(filepath, "rb")),
            #     data=header_data,
            # )

            if ret.status_code != 200:  # pragma: no cover
                raise colrev_exceptions.TEIException()

            if b"[TIMEOUT]" in ret.content:  # pragma: no cover
                raise colrev_exceptions.TEITimeoutException()

            self.root = etree.fromstring(ret.content)

            if self.tei_path is not None:
                self.tei_path.parent.mkdir(exist_ok=True, parents=True)
                with open(self.tei_path, "wb") as file:
                    file.write(ret.content)

                # Note : reopen/write to prevent format changes in the enhancement
                with open(self.tei_path, "rb") as file:
                    xml_fstring = file.read()
                self.root = etree.fromstring(xml_fstring)

                tree = etree.ElementTree(self.root)
                tree.write(str(self.tei_path), encoding="utf-8")
        except requests.exceptions.ConnectionError as exc:  # pragma: no cover
            print(exc)
            print(str(self.pdf_path))
            raise colrev_exceptions.TEITimeoutException() from exc

    def get_tei_str(self) -> str:
        """Get the TEI string"""
        try:
            etree.register_namespace("tei", "http://www.tei-c.org/ns/1.0")
            return etree.tostring(self.root).decode("utf-8")
        except XMLSyntaxError as exc:  # pragma: no cover
            raise colrev_exceptions.TEIException from exc

    def get_grobid_version(self) -> str:
        """Get the GROBID version used for TEI creation"""
        grobid_version = ""
        encoding_description = self.root.find(".//" + self.ns["tei"] + "encodingDesc")
        if encoding_description is not None:
            app_info_node = encoding_description.find(
                ".//" + self.ns["tei"] + "appInfo"
            )
            if app_info_node is not None:
                application_node = encoding_description.find(
                    ".//" + self.ns["tei"] + "application"
                )
                if application_node is not None:
                    if application_node.get("version") is not None:
                        grobid_version = application_node.get("version")
        return grobid_version

    def _parse_author_dict(self, *, author_pers_node: Element) -> dict:
        author_dict = {}
        surname_node = author_pers_node.find(self.ns["tei"] + "surname")
        if surname_node is not None:
            surname = surname_node.text if surname_node.text is not None else ""
            author_dict["surname"] = surname
        else:
            author_dict["surname"] = ""

        forename_node = author_pers_node.find(
            self.ns["tei"] + 'forename[@type="first"]'
        )
        if forename_node is not None:
            forename = forename_node.text if forename_node.text is not None else ""
            if 1 == len(forename):
                forename = forename + "."
            author_dict["forename"] = forename
        else:
            author_dict["forename"] = ""

        middlename_node = author_pers_node.find(
            self.ns["tei"] + 'forename[@type="middle"]'
        )
        if middlename_node is not None:
            middlename = (
                " " + middlename_node.text if middlename_node.text is not None else ""
            )
            author_dict["middlename"] = middlename
        else:
            middlename = ""

        return author_dict

    def _get_author_name_from_node(self, *, author_node: Element) -> str:
        authorname = ""

        author_pers_node = author_node.find(self.ns["tei"] + "persName")
        if author_pers_node is None:
            return authorname

        author_dict = self._parse_author_dict(author_pers_node=author_pers_node)

        authorname = author_dict["surname"] + ", " + author_dict["forename"]
        if "middlename" in author_dict:
            authorname += author_dict["middlename"]

        authorname = (
            authorname.replace("\n", " ")
            .replace("\r", "")
            .replace("•", "")
            .replace("+", "")
            .replace("Dipl.", "")
            .replace("Prof.", "")
            .replace("Dr.", "")
            .replace("&apos", "'")
            .replace("❚", "")
            .replace("~", "")
            .replace("®", "")
            .replace("|", "")
        )

        authorname = re.sub("^Paper, Short; ", "", authorname)
        return authorname

    def _get_paper_doi(self, reference: Element) -> str:
        doi = ""
        dois = reference.findall(".//" + self.ns["tei"] + "idno[@type='DOI']")
        for res in dois:
            if res.text is not None:
                doi = res.text
        return doi

    def get_abstract(self) -> str:
        """Get the abstract"""

        html_tag_regex = re.compile("<.*?>")

        def cleanhtml(raw_html: str) -> str:
            cleantext = re.sub(html_tag_regex, "", raw_html)
            return cleantext

        abstract_text = ""
        profile_description = self.root.find(".//" + self.ns["tei"] + "profileDesc")
        if profile_description is not None:
            abstract_node = profile_description.find(
                ".//" + self.ns["tei"] + "abstract"
            )
            html_str = etree.tostring(abstract_node).decode("utf-8")
            abstract_text = cleanhtml(html_str)
        abstract_text = abstract_text.lstrip().rstrip()
        return abstract_text

    def get_metadata(self) -> dict:
        """Get the metadata of the PDF (title, author, ...) as a dict"""

        reference = self.root.find(".//" + self.ns["tei"] + "sourceDesc").find(
            ".//" + self.ns["tei"] + "biblStruct"
        )
        record = self._get_dict_from_reference(reference)

        if Fields.TITLE in record:
            record[Fields.TITLE] = (
                record[Fields.TITLE]
                .replace("(Completed paper)", "")
                .replace("(Completed-paper)", "")
                .replace("(Research-in-Progress)", "")
                .replace("Completed Research Paper", "")
            )

        for key, value in record.items():
            if key != Fields.FILE:
                record[key] = value.replace("}", "").replace("{", "").rstrip("\\")

        record = {k: v for k, v in record.items() if v != ""}
        return record

    def get_paper_keywords(self) -> list:
        """Get hte keywords"""
        keywords = []
        for keyword_list in self.root.iter(self.ns["tei"] + "keywords"):
            for keyword in keyword_list.iter(self.ns["tei"] + "term"):
                keywords.append(keyword.text)
        return keywords

    def get_author_details(self) -> list:
        """Get the author details"""
        author_details = []

        file_description = self.root.find(".//" + self.ns["tei"] + "sourceDesc")

        if file_description is not None:
            if file_description.find(".//" + self.ns["tei"] + "analytic") is not None:
                analytic_node = file_description.find(
                    ".//" + self.ns["tei"] + "analytic"
                )
                if analytic_node is not None:
                    for author_node in analytic_node.iterfind(
                        self.ns["tei"] + "author"
                    ):
                        author_pers_node = author_node.find(self.ns["tei"] + "persName")
                        if author_pers_node is None:
                            continue

                        author_dict = self._parse_author_dict(
                            author_pers_node=author_pers_node
                        )

                        email_node = author_node.find(self.ns["tei"] + "email")
                        if email_node is not None:
                            author_dict["emai"] = email_node.text

                        orcid_node = author_node.find(
                            self.ns["tei"] + 'idno[@type="ORCID"]'
                        )
                        if orcid_node is not None:
                            orcid = orcid_node.text
                            author_dict["ORCID"] = orcid

                        author_details.append(author_dict)

        return author_details

    # reference elements  ----------------------------

    def _get_reference_bibliography_tei_id(self, reference: Element) -> str:
        tei_id = ""
        try:
            tei_id = reference.attrib[self.ns["w3"] + "id"]
        except KeyError:
            pass
        return tei_id

    def _get_reference_author_string(self, reference: Element) -> str:
        author_list = []
        if reference.find(self.ns["tei"] + "analytic") is not None:
            authors_node = reference.find(self.ns["tei"] + "analytic")
        elif reference.find(self.ns["tei"] + "monogr") is not None:
            authors_node = reference.find(self.ns["tei"] + "monogr")
        else:
            return ""

        if authors_node is not None:
            for author_node in authors_node.iterfind(self.ns["tei"] + "author"):
                authorname = self._get_author_name_from_node(author_node=author_node)

                if authorname not in [", ", ""]:
                    author_list.append(authorname)

        author_string = " and ".join(author_list)

        author_string = (
            author_string.replace("\n", " ")
            .replace("\r", "")
            .replace("•", "")
            .replace("+", "")
            .replace("Dipl.", "")
            .replace("Prof.", "")
            .replace("Dr.", "")
            .replace("&apos", "'")
            .replace("❚", "")
            .replace("~", "")
            .replace("®", "")
            .replace("|", "")
        )
        return author_string

    def _get_reference_title_string(self, reference: Element) -> str:
        title_string = ""
        title = None
        if reference.find(self.ns["tei"] + "analytic") is not None:
            analytic_node = reference.find(self.ns["tei"] + "analytic")
            if analytic_node is not None:
                title = analytic_node.find(self.ns["tei"] + "title")
        elif reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                title = monogr_node.find(self.ns["tei"] + "title")

        if title is not None:
            if title.text is not None:
                title_string = title.text

        return title_string

    def _get_reference_year_string(self, reference: Element) -> str:
        year_string = ""
        year = None
        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                imprint_node = monogr_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    year = imprint_node.find(self.ns["tei"] + "date")

        if year is not None:
            for name, value in sorted(year.items()):
                year_string = value if (name == "when") else ""

        match = re.match(r"\d{4}", year_string)
        year_string = match.group(0) if match else year_string
        return year_string

    def _get_reference_page_string(self, reference: Element) -> str:
        page_string = ""
        page_list = []

        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                imprint_node = monogr_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    page_list = imprint_node.findall(
                        self.ns["tei"] + "biblScope[@unit='page']"
                    )

        for page in page_list:
            if page is not None:
                for name, value in sorted(page.items()):
                    if name == "from":
                        page_string += value
                    if name == "to":
                        page_string += "--" + value

        return page_string

    def _get_reference_number_string(self, reference: Element) -> str:
        number_string = ""
        number_list = []

        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                imprint_node = monogr_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    number_list = imprint_node.findall(
                        self.ns["tei"] + "biblScope[@unit='issue']"
                    )

        for number in number_list:
            if number.text is not None:
                number_string = number.text

        return number_string

    def _get_reference_volume_string(self, reference: Element) -> str:
        volume_string = ""
        volume_list = []

        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                imprint_node = monogr_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    volume_list = imprint_node.findall(
                        self.ns["tei"] + "biblScope[@unit='volume']"
                    )

        for volume in volume_list:
            if volume.text is not None:
                volume_string = volume.text

        return volume_string

    def _get_reference_monograph_string(self, reference: Element) -> str:
        journal_title = ""
        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                monogr_title = monogr_node.find(self.ns["tei"] + "title")
                if monogr_title is not None:
                    if monogr_title.text is not None:
                        journal_title = monogr_title.text

        return journal_title

    # pylint: disable=too-many-nested-blocks
    def _get_entrytype(self, reference: Element) -> str:
        entrytype = ENTRYTYPES.MISC
        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                meeting_node = monogr_node.find(self.ns["tei"] + "meeting")

                if meeting_node is not None:
                    entrytype = ENTRYTYPES.INPROCEEDINGS
                else:
                    title_node = monogr_node.find(self.ns["tei"] + "title")
                    if title_node is not None:
                        if title_node.text is not None:
                            if any(
                                x in title_node.text.lower()
                                for x in [
                                    "conference",
                                    "proceedings",
                                    "symposium",
                                    "meeting",
                                ]
                            ):
                                entrytype = ENTRYTYPES.INPROCEEDINGS
                                return entrytype

                        if title_node.get("level", "") == "j":
                            entrytype = ENTRYTYPES.ARTICLE
                        elif title_node.get("level", "") in ["m", "s"]:
                            entrytype = ENTRYTYPES.BOOK
        return entrytype

    def _get_tei_id_count(self, *, tei_id: str) -> int:
        count = 0

        for reference in self.root.iter(self.ns["tei"] + "ref"):
            if "target" in reference.keys():
                if reference.get("target") == f"#{tei_id}":
                    count += 1

        return count

    def _get_dict_from_reference(self, reference: Element) -> dict:
        entrytype = self._get_entrytype(reference)
        tei_id = self._get_reference_bibliography_tei_id(reference)

        ground_dict = {
            Fields.ID: tei_id,
            Fields.TEI_ID: tei_id,
            Fields.AUTHOR: self._get_reference_author_string(reference),
            Fields.TITLE: self._get_reference_title_string(reference),
            Fields.YEAR: self._get_reference_year_string(reference),
            Fields.PAGES: self._get_reference_page_string(reference),
            Fields.DOI: self._get_paper_doi(reference),
        }

        if entrytype == ENTRYTYPES.ARTICLE:
            ref_rec = {
                Fields.ENTRYTYPE: entrytype,
                Fields.JOURNAL: self._get_reference_monograph_string(reference),
                Fields.VOLUME: self._get_reference_volume_string(reference),
                Fields.NUMBER: self._get_reference_number_string(reference),
            }
        elif entrytype == ENTRYTYPES.BOOK:
            ref_rec = {
                Fields.ENTRYTYPE: entrytype,
                Fields.TITLE: self._get_reference_monograph_string(reference),
            }
        elif entrytype == ENTRYTYPES.INPROCEEDINGS:
            ref_rec = {
                Fields.ENTRYTYPE: entrytype,
                Fields.BOOKTITLE: self._get_reference_monograph_string(reference),
            }
        else:
            ref_rec = {
                Fields.ENTRYTYPE: ENTRYTYPES.MISC,
            }
        ref_rec = {**ground_dict, **ref_rec}

        ref_rec = {k: v for k, v in ref_rec.items() if v != ""}
        return ref_rec

    def get_references(self, *, add_intext_citation_count: bool = False) -> list:
        """Get the bibliography (references section) as a list of record dicts"""
        # Note : could also allow top-10 % of most frequent in-text citations

        #  https://epidoc.stoa.org/gl/latest/ref-title.html

        bibliographies = self.root.iter(self.ns["tei"] + "listBibl")
        tei_bib_db = []
        for bibliography in bibliographies:
            for reference in bibliography:
                ref_rec = self._get_dict_from_reference(reference)

                if add_intext_citation_count:
                    nr_citations = self._get_tei_id_count(tei_id=ref_rec[Fields.ID])
                    ref_rec[Fields.NR_INTEXT_CITATIONS] = nr_citations  # type: ignore

                tei_bib_db.append(ref_rec)

        return tei_bib_db

    def get_citations_per_section(self) -> dict:
        """Get a dict of section-names and list-of-citations"""
        section_citations = {}
        parent_map = {c: p for p in self.root.iter() for c in p}
        sections = self.root.iter(f'{self.ns["tei"]}head')
        for section in sections:
            section_name = section.text
            if section_name is not None:
                parent = parent_map[section]
                citation_nodes = parent.findall(f'.//{self.ns["tei"]}ref')
                citations = [
                    x.get("target", "").replace("#", "")
                    for x in citation_nodes
                    if x.get("type", "") == "bibr"
                ]
                citations = list(filter(lambda a: a != "", citations))
                if len(citations) > 0:
                    section_citations[section_name.lower()] = citations
        return section_citations

    def mark_references(self, *, records: dict):  # type: ignore
        """Mark references with the additional record ID"""

        tei_records = self.get_references()
        for record_dict in tei_records:
            if Fields.TITLE not in record_dict:
                continue

            max_sim = 0.9
            max_sim_record = {}
            for local_record_dict in records.values():
                if local_record_dict[Fields.STATUS] not in [
                    RecordState.rev_included,
                    RecordState.rev_synthesized,
                ]:
                    continue
                rec_sim = colrev.record.record.Record.get_record_similarity(
                    colrev.record.record.Record(record_dict),
                    colrev.record.record.Record(local_record_dict),
                )
                if rec_sim > max_sim:
                    max_sim_record = local_record_dict
                    max_sim = rec_sim
            if len(max_sim_record) == 0:
                continue

            # Record found: mark in tei
            bibliography = self.root.find(f".//{self.ns['tei']}listBibl")
            # mark reference in bibliography
            for ref in bibliography:
                if ref.get(f'{self.ns["w3"]}id') == record_dict[Fields.TEI_ID]:
                    ref.set(Fields.ID, max_sim_record[Fields.ID])
            # mark reference in in-text citations
            for reference in self.root.iter(f'{self.ns["tei"]}ref'):
                if "target" in reference.keys():
                    if reference.get("target") == f"#{record_dict['tei_id']}":
                        reference.set(Fields.ID, max_sim_record[Fields.ID])

            # if settings file available: dedupe_io match agains records

        if self.tei_path:
            tree = etree.ElementTree(self.root)
            tree.write(str(self.tei_path))

        return self.root



#! /usr/bin/env python
"""Narrative review"""
import zope.interface
from pydantic import Field

import colrev.ops.search
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code
# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.ReviewTypeInterface)
class NarrativeReview:
    """Narrative review"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=True)

    def __init__(
        self, *, operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.settings = self.settings_class(**settings)

    def __str__(self) -> str:
        return "narrative review"

    def initialize(
        self, settings: colrev.settings.Settings
    ) -> colrev.settings.Settings:
        """Initialize a narrative review"""

        settings.data.data_package_endpoints = [
            {
                "endpoint": "colrev.paper_md",
                "version": "1.0",
                "word_template": "APA-7.docx",
            }
        ]
        return settings



#! /usr/bin/env python
"""Checker for incomplete fields."""
from __future__ import annotations

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields
from colrev.constants import FieldValues

# pylint: disable=too-few-public-methods


class IncompleteFieldChecker:
    """The IncompleteFieldChecker"""

    msg = DefectCodes.INCOMPLETE_FIELD

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the missing-field checks"""

        for key in [
            Fields.TITLE,
            Fields.JOURNAL,
            Fields.BOOKTITLE,
            Fields.AUTHOR,
            Fields.ABSTRACT,
        ]:
            if (
                self._institutional_author(key=key, record=record)
                or record.data.get(key, FieldValues.UNKNOWN) == FieldValues.UNKNOWN
                or record.ignored_defect(key=key, defect=self.msg)
            ):
                record.remove_field_provenance_note(key=key, note=self.msg)
                continue

            if record.masterdata_is_curated():
                continue

            if self._incomplete_field(record=record, key=key):
                record.add_field_provenance_note(key=key, note=self.msg)
            else:
                record.remove_field_provenance_note(key=key, note=self.msg)

    def _incomplete_field(
        self, *, record: colrev.record.record.Record, key: str
    ) -> bool:
        """check for incomplete field"""
        if record.data[key].endswith("...") or record.data[key].endswith("…"):
            return True
        if key == Fields.AUTHOR:
            if (
                # heuristics for missing first names:
                ", and " in record.data[key]
                or record.data[key].rstrip().endswith(",")
                or "," not in record.data[key]
            ):
                return True
        return False

    def _institutional_author(
        self, *, key: str, record: colrev.record.record.Record
    ) -> bool:
        if key != Fields.AUTHOR or Fields.AUTHOR not in record.data:
            return False
        if record.data[Fields.AUTHOR].startswith("{") and record.data[
            Fields.AUTHOR
        ].endswith("}"):
            return True
        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(IncompleteFieldChecker(quality_model))



#!/usr/bin/env python3
"""Searchsource:OSF"""
from __future__ import annotations

import typing
from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.env.environment_manager
import colrev.loader.load_utils
import colrev.ops.load
import colrev.ops.prep
import colrev.ops.search
import colrev.ops.search_api_feed
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.packages.osf.src.osf_api
import colrev.process.operation
import colrev.record.record
import colrev.record.record_prep
import colrev.review_manager
import colrev.settings
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.osf.src.osf_api import OSFApiQuery

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class OSFSearchSource:
    """OSF"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    source_identifier = Fields.ID
    search_types = [SearchType.API]
    endpoint = "colrev.osf"

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.oni

    db_url = "https://osf.io/"
    SETTINGS = {
        "api_key": "packages.search_source.colrev.osf.api_key",
    }

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager

        if settings:
            self.search_source = self.settings_class(**settings)
        else:
            self.search_source = colrev.settings.SearchSource(
                endpoint=self.endpoint,
                filename=Path("data/search/osf.bib"),
                search_type=SearchType.API,
                search_parameters={},
                comment="",
            )
            self.source_operation = source_operation

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for OSF"""

        result = {"confidence": 0.1}

        return result

    @classmethod
    def add_endpoint(
        cls, operation: colrev.ops.search.Search, params: dict
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a)"""

        params_dict: typing.Dict[str, str] = {}
        if params and isinstance(params, str) and params.startswith("http"):
            params_dict = {Fields.URL: params}

        # Select the search type based on the provided parameters
        search_type = operation.select_search_type(
            search_types=cls.search_types, params={"query": params_dict}
        )

        # Handle different search types
        if search_type == SearchType.API:
            # Check for params being empty and initialize if needed
            if len(params_dict) == 0:
                search_source = operation.create_api_source(endpoint=cls.endpoint)
                # Search title per default (other fields may be supported later)
                search_source.search_parameters["query"] = {
                    "title": search_source.search_parameters["query"]
                }
            elif "https://api.osf.io/v2/nodes/?filter" in params_dict.get("url", ""):
                query = (
                    params_dict["url"]
                    .replace("https://api.osf.io/v2/nodes/?filter", "")
                    .lstrip("&")
                )

                parameter_pairs = query.split("&")
                search_parameters = {
                    key.lstrip("[").rstrip("]"): value
                    for key, value in (pair.split("=") for pair in parameter_pairs)
                }
                last_value = list(search_parameters.values())[-1]
                filename = operation.get_unique_filename(
                    file_path_string=f"osf_{last_value}"
                )
                search_source = colrev.settings.SearchSource(
                    endpoint=cls.endpoint,
                    filename=filename,
                    search_type=SearchType.API,
                    search_parameters={"query": search_parameters},
                    comment="",
                )
        else:
            raise NotImplementedError("Unsupported search type.")

        # Adding the source and performing the search
        operation.add_source_and_search(search_source)
        return search_source

    def _get_api_key(self) -> str:
        api_key = self.review_manager.environment_manager.get_settings_by_key(
            self.SETTINGS["api_key"]
        )
        if api_key is None or len(api_key) == 0:
            api_key = input("Please enter api key: ")
            self.review_manager.environment_manager.update_registry(
                self.SETTINGS["api_key"], api_key
            )
        return api_key

    def search(self, rerun: bool) -> None:
        """Run a search of OSF"""
        osf_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )
        self._run_api_search(osf_feed=osf_feed, rerun=rerun)

    def _run_api_search(
        self, osf_feed: colrev.ops.search_api_feed.SearchAPIFeed, rerun: bool
    ) -> None:

        api = OSFApiQuery(
            parameters=self.search_source.search_parameters["query"],
            api_key=self._get_api_key(),
        )
        self.review_manager.logger.info(f"Retrieve {api.overall()} records")

        while True:
            for record_dict in api.retrieve_records():
                record = colrev.record.record.Record(record_dict)
                osf_feed.add_update_record(record)

            if api.pages_completed():
                break

        osf_feed.save()

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 60,
    ) -> colrev.record.record.Record:
        """Not implemented"""

        return record

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
        source: colrev.settings.SearchSource,
    ) -> colrev.record.record.Record:
        """Needs manual preparation"""

        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records."""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=load_operation.review_manager.logger,
                unique_id_field="ID",
            )
            return records

        raise NotImplementedError



#!/usr/bin/env python3
"""Merge branches of CoLRev projects."""
from __future__ import annotations

import copy

from dictdiffer import diff
from git.exc import GitCommandError

import colrev.process.operation
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import OperationsType

# pylint: disable=too-few-public-methods


class Merge(colrev.process.operation.Operation):
    """Merge branches of CoLRev project"""

    type = OperationsType.check

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=False,
        )

    def _get_non_status_changes(
        self,
        *,
        current_branch_records: dict,
        other_branch_records: dict,
        current_branch_name: str,
        other_branch_name: str,
    ) -> list:
        non_status_changes = []

        records_missing_in_current_branch = [
            rid for rid in other_branch_records if rid not in current_branch_records
        ]
        if records_missing_in_current_branch:
            non_status_changes.append(
                {
                    f"records missing in {current_branch_name}": records_missing_in_current_branch
                }
            )

        records_missing_in_other_branch = [
            rid for rid in current_branch_records if rid not in other_branch_records
        ]
        if records_missing_in_other_branch:
            non_status_changes.append(
                {
                    f"records missing in {other_branch_name}:": records_missing_in_other_branch
                }
            )

        changed_records = []
        for current_record_id, current_record in current_branch_records.items():
            if current_record_id not in other_branch_records:
                continue
            other_record = other_branch_records[current_record_id]

            comparison_current_record = current_record.copy()
            del comparison_current_record[Fields.STATUS]
            comparison_other_record = other_record.copy()
            del comparison_other_record[Fields.STATUS]

            if comparison_current_record != comparison_other_record:
                comparison_diff = self.review_manager.p_printer.pformat(
                    list(diff(comparison_current_record, comparison_other_record))
                )
                changed_records.append(f"{current_record_id}: " f"{comparison_diff}")
        if changed_records:
            non_status_changes.append({"changed record fields:": changed_records})

        return non_status_changes

    @colrev.process.operation.Operation.decorate()
    def main(self, *, branch: str) -> None:
        """Merge branches of a CoLRev project (main entrypoint)"""

        # pylint: disable=too-many-locals
        # pylint: disable=too-many-statements

        git_repo = self.review_manager.dataset.get_repo()
        # our_index  = git_repo.index

        for remote in git_repo.remotes:
            remote.fetch()

        branches = git_repo.heads
        assert branch in [b.name for b in branches]

        git_branch = [b for b in branches if b.name == branch][0]
        merging_branch_author = git_branch.commit.author
        current_branch = git_repo.active_branch.name

        try:
            git_repo.git.merge(branch)
            self.review_manager.logger.info("Merged without conflicts.")
            return
        except GitCommandError:
            self.review_manager.logger.info("Detected changes in both branches.")

        unmerged_blobs = git_repo.index.unmerged_blobs()

        # Note : only two-way merges supported for now.
        assert all(len(v) == 3 for k, v in unmerged_blobs.items())

        # Ensure the path uses forward slashes, which is compatible with Git's path handling
        if self.review_manager.paths.RECORDS_FILE_GIT in unmerged_blobs:
            current_branch_records = {}
            other_branch_records = {}
            for stage, blob in unmerged_blobs[
                self.review_manager.paths.RECORDS_FILE_GIT
            ]:
                # stage == 1: common ancestor (often md_processed for prescreen)
                # stage == 2: own branch
                # stage == 3: other branch
                if 2 == stage:

                    current_branch_records = colrev.loader.load_utils.loads(
                        load_string=blob.data_stream.read().decode("utf-8"),
                        implementation="bib",
                        logger=self.review_manager.logger,
                    )

                elif 3 == stage:
                    other_branch_records = colrev.loader.load_utils.loads(
                        load_string=blob.data_stream.read().decode("utf-8"),
                        implementation="bib",
                        logger=self.review_manager.logger,
                    )

        else:
            self.review_manager.logger.info(
                f"No conflicts to reconcile in {self.review_manager.paths.records}."
            )
            return

        # There may be removed records / renamed IDs, changed fields...
        # if so: print, ask to resolve and exit
        non_status_changes = self._get_non_status_changes(
            current_branch_records=current_branch_records,
            other_branch_records=other_branch_records,
            current_branch_name=current_branch,
            other_branch_name=branch,
        )
        if non_status_changes:
            print(
                "Resolve non-status changes before merging "
                "(abort merge using git merge --abort):"
            )
            print(non_status_changes)
            return

        self.review_manager.logger.info("Reconciling changes in colrev_status.")
        # Note : reconciliation of other changes not supported yet

        (
            current_branch_author,
            _,
        ) = self.review_manager.environment_manager.get_name_mail_from_git()

        self.review_manager.logger.info(
            "Start merge reconciliation: "
            f"branch {current_branch} ({current_branch_author}) <-> "
            f"branch {branch} ({merging_branch_author})"
        )

        # Copy: for statistics
        current_branch_records_prior = copy.deepcopy(current_branch_records)

        print()
        nr_to_reconcile = len(
            [
                r
                for r in current_branch_records.values()
                if other_branch_records[r[Fields.ID]][Fields.STATUS] != r[Fields.STATUS]
            ]
        )
        i = 0
        for (
            current_branch_record_id,
            current_branch_record_dict,
        ) in current_branch_records.items():
            other_branch_record = other_branch_records[current_branch_record_id]

            if (
                current_branch_record_dict[Fields.STATUS]
                != other_branch_record[Fields.STATUS]
            ):
                i += 1
                print(f"{i}/{nr_to_reconcile}")
                copied_rec = current_branch_record_dict.copy()
                copied_rec.pop(Fields.STATUS)
                print(colrev.record.record.Record(copied_rec).format_bib_style())
                print(
                    f"1 - {current_branch_author} coded on {current_branch}".ljust(
                        40, " "
                    )
                    + f": {current_branch_record_dict['colrev_status']}"
                )
                print(
                    f"2 - {merging_branch_author} coded on {branch}".ljust(40, " ")
                    + f": {other_branch_record['colrev_status']}"
                )
                resolution_nr = input("Enter resolution: (1 or 2)")
                if resolution_nr == "1":
                    self.review_manager.report_logger.info(
                        f"Reconciliation for {current_branch_record_id}: "
                        f"{current_branch_record_dict['colrev_status']}"
                    )
                    resolution = current_branch_record_dict[Fields.STATUS]
                else:
                    self.review_manager.report_logger.info(
                        f"Reconciliation for {current_branch_record_id}: "
                        f"{other_branch_record['colrev_status']}"
                    )
                    resolution = other_branch_record[Fields.STATUS]
                current_branch_record = colrev.record.record.Record(
                    current_branch_record_dict
                )
                current_branch_record.set_status(resolution)
                print("\n\n\n")

        self.review_manager.dataset.save_records_dict(current_branch_records)

        self.review_manager.update_status_yaml(add_to_git=False)

        validate_operation = self.review_manager.get_validate_operation()
        validation_details = validate_operation.validate_merge_prescreen_screen(
            current_branch_records=current_branch_records_prior,
            other_branch_records=other_branch_records,
            records_reconciled=current_branch_records,
        )
        print("Statistics:")
        self.review_manager.p_printer.pprint(validation_details["statistics"])

        print(
            f"\n{Colors.ORANGE}Please add (git add .) and commit (git commit){Colors.END}"
        )

        # Note : cannot add/create commit yet - not yet supported by gitpython:
        # https://github.com/gitpython-developers/GitPython/issues/1185
        # our_index.write(ignore_extension_data=True)



#! /usr/bin/env python3
"""Synchronize records into a non-CoLRev project."""
from __future__ import annotations

import logging
import re
import typing
from pathlib import Path

import pybtex.errors

import colrev.dataset
import colrev.env.local_index
import colrev.env.local_index_builder
import colrev.exceptions as colrev_exceptions
import colrev.loader.load_utils_formatter
import colrev.ops.check
import colrev.package_manager.package_manager
import colrev.record.record
import colrev.record.record_similarity
import colrev.review_manager
import colrev.ui_cli.cli_status_printer
import colrev.ui_cli.cli_validation
import colrev.ui_cli.dedupe_errors
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import RecordState
from colrev.packages.crossref.src import crossref_api
from colrev.writer.write_utils import write_file


class Sync:
    """Synchronize records into a non-CoLRev repository"""

    cited_papers: list
    pre_commit_config = Path(".pre-commit-config.yaml")
    references_file = Path("references.bib")

    def __init__(self) -> None:
        self.records_to_import: typing.List[colrev.record.record.Record] = []
        self.non_unique_for_import: typing.List[typing.Dict] = []

        self.logger = self._setup_logger(level=logging.DEBUG)
        self.paper_md = self._get_md_file()
        self.load_formatter = colrev.loader.load_utils_formatter.LoadFormatter()

    def add_hook(self) -> None:
        """Add a pre-commit hook for colrev sync"""
        if not Path(".git").is_dir():
            print("Not in a git directory.")
            return
        if not Path("records.bib").is_file() or not Path("paper.md").is_file():
            print("Warning: records.bib or paper.md does not exist.")
            print("Other filenames are not (yet) supported.")
            return

        if self.pre_commit_config.is_file():
            if "colrev-hooks-update" in self.pre_commit_config.read_text(
                encoding="utf-8"
            ):
                print("Hook already registered")
                return

        with open(self.pre_commit_config, "a", encoding="utf-8") as file:
            file.write(
                """\n-   repo: local
        hooks:
        -   id: colrev-hooks-update
            name: "CoLRev ReviewManager: update"
            entry: colrev-hooks-update
            language: python
            stages: [commit]
            files: 'records.bib|paper.md'"""
            )
        print("Added pre-commit hook for colrev sync.")

    def _setup_logger(self, *, level: int = logging.INFO) -> logging.Logger:
        """Setup the sync logger"""
        # pylint: disable=duplicate-code

        # for logger debugging:
        # from logging_tree import printout
        # printout()
        logger = logging.getLogger(f"colrev{str(Path.cwd()).replace('/', '_')}")
        logger.setLevel(level)

        if logger.handlers:
            for handler in logger.handlers:
                logger.removeHandler(handler)

        formatter = logging.Formatter(
            fmt="%(asctime)s [%(levelname)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        handler.setLevel(level)

        logger.addHandler(handler)
        logger.propagate = False

        return logger

    def _get_md_file(self) -> Path:
        paper_md = Path("")
        if Path("paper.md").is_file():
            paper_md = Path("paper.md")
        if Path("data/paper.md").is_file():
            paper_md = Path("data/paper.md")
        elif Path("review.md").is_file():
            paper_md = Path("review.md")
        return paper_md

    def _get_cited_papers_citation_keys(self) -> list:
        rst_files = list(Path.cwd().rglob("*.rst"))

        citation_keys = []
        if self.paper_md.is_file():
            self.logger.info("Load cited references from paper.md")
            content = self.paper_md.read_text(encoding="utf-8")
            res = re.findall(r"(^|\s|\[|;)(@[a-zA-Z0-9_]+)+", content)
            citation_keys = list({r[1].replace("@", "") for r in res})
            if "fig" in citation_keys:
                citation_keys.remove("fig")
            self.logger.info("Citations in paper.md: %s", len(citation_keys))
            self.cited_papers = citation_keys

        elif len(rst_files) > 0:
            self.logger.info("Load cited references from *.rst")
            for rst_file in rst_files:
                content = rst_file.read_text()
                res = re.findall(r":cite:p:`(.*)`", content)
                cited = [c for cit_group in res for c in cit_group.split(",")]
                citation_keys.extend(cited)

            citation_keys = list(set(citation_keys))
            self.logger.info("Citations in *.rst: %s", len(citation_keys))
            self.cited_papers = citation_keys

        else:
            print("Not found paper.md or *.rst")
        return citation_keys

    def get_cited_papers_from_source(self, *, src: Path) -> None:
        """Get the cited papers from a source file"""

        citation_keys = self._get_cited_papers_citation_keys()

        ids_in_bib = self._get_ids_in_paper()
        self.logger.info("References in bib: %s", len(ids_in_bib))

        if src.suffix == ".bib":
            refs_in_src = colrev.loader.load_utils.load(
                filename=src,
                logger=self.logger,
            )

        else:
            print("Format not supported")
            return

        for citation_key in citation_keys:
            if citation_key in ids_in_bib:
                continue
            if citation_key not in refs_in_src:
                print(f"{citation_key} not in {src}")
                continue
            self.records_to_import.append(
                colrev.record.record.Record(refs_in_src[citation_key])
            )

    def get_cited_papers(self) -> None:
        """Get the cited papers"""

        citation_keys = self._get_cited_papers_citation_keys()

        ids_in_bib = self._get_ids_in_paper()
        self.logger.info("References in bib: %s", len(ids_in_bib))

        local_index = colrev.env.local_index.LocalIndex()
        for citation_key in citation_keys:
            if citation_key in ids_in_bib + ["tbl"]:
                continue

            if Path(f"{citation_key}.pdf").is_file():
                print("TODO - prefer!")
                # continue if found/extracted

            returned_records = local_index.search(f"citation_key='{citation_key}'")

            if 0 == len(returned_records):
                self.logger.info("Not found: %s", citation_key)
            elif 1 == len(returned_records):
                if returned_records[0].data[Fields.ID] in [
                    r.data[Fields.ID] for r in self.records_to_import
                ]:
                    continue
                self.records_to_import.append(returned_records[0])
            else:
                listed_item: typing.Dict[str, typing.List] = {citation_key: []}
                for returned_record in returned_records:  # type: ignore
                    listed_item[citation_key].append(returned_record)
                self.non_unique_for_import.append(listed_item)

    def _get_ids_in_paper(self) -> typing.List:
        pybtex.errors.set_strict_mode(False)

        if self.references_file.is_file():

            records = colrev.loader.load_utils.load(
                filename=self.references_file,
                logger=self.logger,
            )

        else:
            records = {}

        return list(records.keys())

    def add_to_records_to_import(self, record: colrev.record.record.Record) -> None:
        """Add a record to the records_to_import list"""
        if record.data[Fields.ID] not in [
            r.data[Fields.ID] for r in self.records_to_import
        ]:
            self.records_to_import.append(record)

    def add_paper(self, add: str) -> None:
        """Add a paper to the bibliography"""

        local_index = colrev.env.local_index.LocalIndex()

        def parse_record_str(add: str) -> list:
            # DOI: from crossref
            if add.startswith("10."):
                returned_records = local_index.search(f"doi='{add}'")
            else:
                returned_records = local_index.search(f"title LIKE '{add}'")

            return returned_records

        self.get_cited_papers()
        records = parse_record_str(add)

        if len(records) == 0:
            print("not found")
            return

        self.records_to_import = records
        input(self.records_to_import)
        print(records[0].data["ID"])

    def add_to_bib(self) -> None:
        """Add records to the bibliography"""

        if not self.paper_md.is_file():
            return

        if self.paper_md.read_text(encoding="utf-8").startswith("---"):
            self._export_to_bib()

        else:
            # Append to # References if no header (mardown with linked)
            self._append_as_citations()

    def _append_as_citations(self) -> None:
        if "# References" in self.paper_md.read_text(encoding="utf-8"):
            print("Already contains a reference section.")
            return

        with open(self.paper_md, "a", encoding="utf-8") as file:
            file.write("\n# References\n\n")
            ref_list = [
                record_to_import.format_bib_style()
                for record_to_import in self.records_to_import
            ]
            file.write("\n".join(sorted(ref_list)))

    def _export_to_bib(self) -> None:
        pybtex.errors.set_strict_mode(False)

        if not self.references_file.is_file():
            records = []
        else:

            records_dict = colrev.loader.load_utils.load(
                filename=self.references_file,
                logger=self.logger,
            )
            records = list(records_dict.values())

        available_ids = [r[Fields.ID] for r in records]
        added = []
        for record_to_import in self.records_to_import:
            if record_to_import.data[Fields.ID] not in available_ids:
                record_to_import = colrev.record.record.Record(
                    {
                        k: v
                        for k, v in record_to_import.data.items()
                        if k
                        not in FieldSet.PROVENANCE_KEYS
                        + [Fields.SCREENING_CRITERIA, Fields.PRESCREEN_EXCLUSION]
                        and "." not in k
                    }
                )
                records.append(record_to_import.get_data())
                available_ids.append(record_to_import.data[Fields.ID])
                added.append(record_to_import)

        if len(added) > 0:
            self.logger.info("Loaded:")
            print()
            for added_record in added:
                added_record.print_citation_format()
            print()

            self.logger.info(
                "%s Loaded %s papers%s", Colors.GREEN, len(added), Colors.END
            )

        for record_dict in records:
            record = colrev.record.record.Record(record_dict)
            record.set_status(RecordState.md_retrieved)  # rev_synthesized
            self.load_formatter.run(record)
            del record.data[Fields.STATUS]

        # records_dict = {
        #     r[Fields.ID]: r for r in records if r[Fields.ID] in self.cited_papers
        # }

        records_dict = {r[Fields.ID]: r for r in records}

        write_file(records_dict=records_dict, filename=self.references_file)

    def udpate_bib(self) -> None:
        """Update the bibliography"""

        if not self.references_file.is_file():
            return

        records = colrev.loader.load_utils.load(
            filename=self.references_file,
            logger=self.logger,
        )

        api = crossref_api.CrossrefAPI(params={})

        for record_dict in records.values():
            if "doi" in record_dict:
                continue
            record = colrev.record.record.Record(record_dict)
            self.logger.info("Retrieve from crossref: %s", record_dict[Fields.ID])

            try:
                retrieved_records = api.crossref_query(
                    record_input=record,
                    jour_vol_iss_list=False,
                )
            except colrev_exceptions.NotEnoughDataToIdentifyException:
                continue
            except colrev_exceptions.ServiceNotAvailableException:
                continue
            if not retrieved_records:
                continue
            retrieved_record = retrieved_records.pop()

            if colrev.record.record_similarity.matches(record, retrieved_record):
                for key, value in retrieved_record.data.items():
                    record.data[key] = value

        write_file(records_dict=records, filename=self.references_file)


def main() -> None:
    """Sync records from CoLRev environment to non-CoLRev repo"""

    sync_operation = Sync()
    sync_operation.udpate_bib()

    sync_operation.get_cited_papers()

    if len(sync_operation.non_unique_for_import) > 0:
        print("Non-unique keys to resolve:")
        # Resolve non-unique cases
        for case in sync_operation.non_unique_for_import:
            for val in case.values():
                # later: there may be more collisions (v3, v4)
                v_1 = val[0].format_bib_style()
                v_2 = val[1].format_bib_style()

                if v_1.lower() == v_2.lower():
                    sync_operation.add_to_records_to_import(val[0])
                    continue
                print("\n")
                print(f"1: {v_1}")
                print("      " + val[0].data.get("source_url", ""))
                print("")
                print(f"2: {v_2}")
                print("      " + val[1].data.get("source_url", ""))
                user_selection = input("Import version 1 or 2 (or skip)?")
                if user_selection == "1":
                    sync_operation.add_to_records_to_import(val[0])
                    continue
                if user_selection == "2":
                    sync_operation.add_to_records_to_import(val[1])
                    continue

    sync_operation.add_to_bib()


# @click.option(
#     "-a",
#     "--add",
#     help="Paper to add.",
#     required=False,
# )
# @click.option(
#     "--add_hook",
#     is_flag=True,
#     default=False,
#     help="Add a sync pre-commit hook",
# )
# @click.option(
#     "-src",
#     type=click.Path(exists=True),
#     help="Sync selected citations from source file.",
# )

# sync_operation = colrev.review_manager.ReviewManager.get_sync_operation()
# if add_hook:
#     sync_operation.add_hook()
#     return

# if src:
#     sync_operation.get_cited_papers_from_source(src=Path(src))
#     sync_operation.add_to_bib()
#     return

# if add:
#     sync_operation.add_paper(add)
#     sync_operation.add_to_bib()

#     return



#! /usr/bin/env python
"""Exclude complementary materials as a prep operation"""
from __future__ import annotations

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields

# pylint: disable=duplicate-code


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class ExcludeComplementaryMaterialsPrep:
    """Prepares records by excluding complementary materials
    (tables of contents, editorial boards, about our authors)"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)

        self.complementary_materials_keywords = (
            colrev.env.utils.load_complementary_material_keywords()
        )

        # for exact matches:
        self.complementary_materials_strings = (
            colrev.env.utils.load_complementary_material_strings()
        )

        # for prefixes:
        self.complementary_material_prefixes = (
            colrev.env.utils.load_complementary_material_prefixes()
        )

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare the records by excluding complementary materials"""

        if (
            any(
                complementary_materials_keyword
                in record.data.get(Fields.TITLE, "").lower()
                for complementary_materials_keyword in self.complementary_materials_keywords
            )
            or any(
                complementary_materials_string
                == record.data.get(Fields.TITLE, "").lower()
                for complementary_materials_string in self.complementary_materials_strings
            )
            or any(
                record.data.get(Fields.TITLE, "").lower().startswith(prefix)
                for prefix in self.complementary_material_prefixes
            )
        ):
            record.prescreen_exclude(reason="complementary material")

        return record



#! /usr/bin/env python
"""Checker for language-unknown."""
from __future__ import annotations

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields
from colrev.constants import FieldValues

# pylint: disable=too-few-public-methods


class LanguageChecker:
    """The LanguageChecker"""

    msg = DefectCodes.LANGUAGE_UNKNOWN

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the language-unknown checks"""

        if record.data.get(
            Fields.TITLE, FieldValues.UNKNOWN
        ) == FieldValues.UNKNOWN or record.ignored_defect(
            key=Fields.TITLE, defect=self.msg
        ):
            return
        if record.masterdata_is_curated():
            return

        if self._language_unknown(record=record):
            record.add_field_provenance_note(key=Fields.TITLE, note=self.msg)

        else:
            record.remove_field_provenance_note(key=Fields.TITLE, note=self.msg)

    def _language_unknown(self, *, record: colrev.record.record.Record) -> bool:
        if Fields.LANGUAGE in record.data:
            return False
        return True


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(LanguageChecker(quality_model))



#! /usr/bin/env python
"""Distribte records or PDFs to local CoLRev repositories."""
from __future__ import annotations

import os
import shutil
from pathlib import Path

import colrev.process.operation
import colrev.settings
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.constants import SearchType
from colrev.writer.write_utils import write_file


class Distribute(colrev.process.operation.Operation):
    """Distribute records to other local CoLRev projects"""

    type = OperationsType.check

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        # pylint: disable=duplicate-code
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=False,
        )
        self.review_manager = review_manager

    def get_environment_registry(self) -> list:
        """Get the environment registry (excluding curated_metadata)"""
        environment_manager = self.review_manager.get_environment_manager()
        return [
            x
            for x in environment_manager.local_repos()
            if "curated_metadata/" not in x["repo_source_path"]
        ]

    def get_next_id(self, *, bib_file: Path) -> int:
        """Get the next ID (incrementing counter)"""
        ids = []
        if bib_file.is_file():
            with open(bib_file, encoding="utf8") as file:
                line = file.readline()
                while line:
                    if "@" in line[:3]:
                        current_id = line[line.find("{") + 1 : line.rfind(",")]
                        ids.append(current_id)
                    line = file.readline()
        max_id = max([int(cid) for cid in ids if cid.isdigit()] + [0]) + 1
        return max_id

    @colrev.process.operation.Operation.decorate()
    def main(self, *, path: Path, target: Path) -> None:
        """Distribute records to other CoLRev repositories (main entrypoint)"""

        # if no options are given, take the current path/repo
        # optional: target-repo-path
        # path_str: could also be a url
        # option: chdir (to target repo)?
        # file: copy or move?

        os.chdir(target)
        path = Path.cwd() / Path(path)

        if path.is_file():
            if path.suffix == ".bib":
                # gh_issue https://github.com/CoLRev-Environment/colrev/issues/69
                # append records (check duplicates/duplicate IDs)
                # if path already exists
                # should the following really rename the file?
                # or just get the updated filepath?

                shutil.move(
                    str(path), str(target / Path("data/search/local_import.bib"))
                )
                input(path)

            if path.suffix == ".pdf":
                grobid_service = self.review_manager.get_grobid_service()

                grobid_service.start()

                tei = self.review_manager.get_tei(
                    pdf_path=path,
                )
                record = tei.get_metadata()

                target_pdf_path = target / "pdfs" / path.name
                target_pdf_path.parent.mkdir(parents=True, exist_ok=True)
                self.review_manager.logger.info(f"Copy PDF to {target_pdf_path}")
                shutil.copyfile(path, target_pdf_path)

                self.review_manager.logger.info(
                    f"append {self.review_manager.p_printer.pformat(record)} "
                    "to data/search/local_import.bib"
                )
                target_bib_file = target / Path("data/search/local_import.bib")
                self.review_manager.logger.info(f"target_bib_file: {target_bib_file}")
                if target_bib_file.is_file():

                    import_records_dict = colrev.loader.load_utils.load(
                        filename=target_bib_file,
                        logger=self.review_manager.logger,
                    )
                    import_records = list(import_records_dict.values())

                else:
                    import_records = []

                    new_source = colrev.settings.SearchSource(
                        endpoint="colrev.unknown_source",
                        filename=Path("search") / target_bib_file.name,
                        search_type=SearchType.OTHER,
                        search_parameters={},
                        comment="",
                    )

                    self.review_manager.settings.sources.append(new_source)
                    self.review_manager.save_settings()

                if 0 == len(import_records):
                    return

                record_id = int(self.get_next_id(bib_file=target_bib_file))
                record[Fields.ID] = f"{record_id}".rjust(10, "0")
                record.update(file=str(target_pdf_path))
                import_records.append(record)

                import_records_dict = {r[Fields.ID]: r for r in import_records}

                write_file(records_dict=import_records_dict, filename=target_bib_file)

                self.review_manager.dataset.add_changes(target_bib_file)



#! /usr/bin/env python
"""Creation of a PRISMA chart as part of the data operations"""
from __future__ import annotations

import os
import typing
from pathlib import Path

import docker
import pandas as pd
import zope.interface
from docker.errors import DockerException
from pydantic import BaseModel
from pydantic import Field

import colrev.env.docker_manager
import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings


@zope.interface.implementer(colrev.package_manager.interfaces.DataInterface)
class PRISMA:
    """Create a PRISMA diagram"""

    ci_supported: bool = Field(default=False)

    class PRISMASettings(
        colrev.package_manager.package_settings.DefaultSettings, BaseModel
    ):
        """PRISMA settings"""

        endpoint: str
        version: str
        diagram_path: typing.List[Path] = Field(
            default_factory=lambda: [Path("PRISMA.png")]
        )

    settings_class = PRISMASettings
    PRISMA_IMAGE = "colrev/prisma:latest"

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.review_manager = data_operation.review_manager
        self.data_operation = data_operation

        # Set default values (if necessary)
        if "version" not in settings:
            settings["version"] = "0.1"

        if "diagram_path" in settings:
            settings["diagram_path"] = [Path(path) for path in settings["diagram_path"]]
        else:
            settings["diagram_path"] = [Path("PRISMA.png")]

        self.settings = self.settings_class(**settings)

        output_dir = self.review_manager.paths.output
        self.csv_path = output_dir / Path("PRISMA.csv")

        self.settings.diagram_path = [
            output_dir / path for path in self.settings.diagram_path
        ]

        if not self.review_manager.in_ci_environment():
            colrev.env.docker_manager.DockerManager.build_docker_image(
                imagename=self.PRISMA_IMAGE
            )
        data_operation.docker_images_to_stop.append(self.PRISMA_IMAGE)

    # pylint: disable=unused-argument
    @classmethod
    def add_endpoint(cls, operation: colrev.ops.data.Data, params: str) -> None:
        """Add as an endpoint"""

        add_package = {
            "endpoint": "colrev.prisma",
            "version": "0.1",
            "diagram_path": ["PRISMA.png"],
        }
        operation.review_manager.settings.data.data_package_endpoints.append(
            add_package
        )
        operation.review_manager.save_settings()
        operation.review_manager.dataset.create_commit(
            msg=f"Add {operation.type} prisma",
        )

    def _export_csv(self, silent_mode: bool) -> None:
        csv_resource_path = Path("packages/prisma/prisma/PRISMA.csv")
        self.csv_path.parent.mkdir(exist_ok=True, parents=True)

        if self.csv_path.is_file():
            os.remove(self.csv_path)
        colrev.env.utils.retrieve_package_file(
            template_file=csv_resource_path, target=self.csv_path
        )

        status_stats = self.review_manager.get_status_stats()

        prisma_data = pd.read_csv(self.csv_path)
        prisma_data["ind"] = prisma_data["data"]
        prisma_data.set_index("ind", inplace=True)
        prisma_data.loc["database_results", "n"] = status_stats.overall.md_retrieved
        prisma_data.loc["duplicates", "n"] = status_stats.md_duplicates_removed
        prisma_data.loc["records_screened", "n"] = status_stats.overall.rev_prescreen
        prisma_data.loc["records_excluded", "n"] = (
            status_stats.overall.rev_prescreen_excluded
        )
        if status_stats.currently.exclusion:
            prisma_data.loc["dbr_excluded", "n"] = "; ".join(
                f"{key}, {val}" for key, val in status_stats.currently.exclusion.items()
            )
        else:
            prisma_data.loc["dbr_excluded", "n"] = (
                f"Overall, {status_stats.overall.rev_excluded}"
            )

        prisma_data.loc["dbr_assessed", "n"] = status_stats.overall.rev_screen
        prisma_data.loc["new_studies", "n"] = status_stats.overall.rev_included
        prisma_data.loc["dbr_notretrieved_reports", "n"] = (
            status_stats.overall.pdf_not_available
        )
        prisma_data.loc["dbr_sought_reports", "n"] = (
            status_stats.overall.rev_prescreen_included
        )

        prisma_data.to_csv(self.csv_path, index=False)
        self.review_manager.logger.debug(f"Exported {self.csv_path}")

        if not status_stats.completeness_condition and not silent_mode:
            self.review_manager.logger.info("Review not (yet) complete")

    def _export_diagram(self, silent_mode: bool) -> None:
        if not self.csv_path.is_file():
            self.review_manager.logger.error("File %s does not exist.", self.csv_path)
            self.review_manager.logger.info("Complete processing and use colrev data")
            return

        csv_relative_path = self.csv_path.relative_to(
            self.review_manager.path
        ).as_posix()

        if not silent_mode:
            self.review_manager.logger.info("Create PRISMA diagram")

        for diagram_path in self.settings.diagram_path:
            diagram_relative_path = diagram_path.relative_to(
                self.review_manager.path
            ).as_posix()

            script = (
                "Rscript "
                + "/prisma.R "
                + f"/data/{csv_relative_path} "
                + f"/data/{diagram_relative_path}"
            )
            # Users can place a custom script in src/prisma.R
            if (self.review_manager.path / Path("src/prisma.R")).is_file():
                script = (
                    "Rscript "
                    + "-e \"source('/data/src/prisma.R')\" "
                    + f"/data/{csv_relative_path} "
                    + f"/data/{diagram_relative_path}"
                )

            self._call_docker_build_process(script=script)
            self.csv_path.unlink()

    def _call_docker_build_process(self, *, script: str) -> None:
        try:
            settings_path = self.review_manager.paths.settings
            uid = os.stat(settings_path).st_uid
            gid = os.stat(settings_path).st_gid
            user = f"{uid}:{gid}"

            client = docker.from_env()

            msg = f"Running docker container created from image {self.PRISMA_IMAGE}"
            self.review_manager.report_logger.info(msg)

            client.containers.run(
                image=self.PRISMA_IMAGE,
                command=script,
                user=user,
                volumes=[os.getcwd() + ":/data"],
                detach=True,
            )
        except docker.errors.ImageNotFound:
            self.review_manager.logger.error("Docker image not found")
        except docker.errors.ContainerError as exc:
            if "Temporary failure in name resolution" in str(exc):
                raise colrev_exceptions.ServiceNotAvailableException(
                    "prisma service failed"
                ) from exc
            raise exc
        except DockerException as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Docker service not available ({exc}). Please install/start Docker."
            ) from exc

    def update_data(
        self,
        records: dict,  # pylint: disable=unused-argument
        synthesized_record_status_matrix: dict,  # pylint: disable=unused-argument
        silent_mode: bool,
    ) -> None:
        """Update the data/prisma diagram"""

        self._export_csv(silent_mode=silent_mode)
        self._export_diagram(silent_mode=silent_mode)

    def update_record_status_matrix(
        self,
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record_status_matrix"""

        # Note : automatically set all to True / synthesized
        for syn_id in list(synthesized_record_status_matrix.keys()):
            synthesized_record_status_matrix[syn_id][endpoint_identifier] = True

    def get_advice(
        self,
    ) -> dict:
        """Get advice on the next steps (for display in the colrev status)"""

        data_endpoint = "Data operation [prisma data endpoint]: "

        path_str = ",".join(
            [
                str(x.relative_to(self.review_manager.path))
                for x in self.settings.diagram_path
            ]
        )
        advice = {
            "msg": f"{data_endpoint}"
            + "\n    - The PRISMA diagram is created automatically "
            + f"({path_str})",
            "detailed_msg": "TODO",
        }
        return advice



#!/usr/bin/env python3
"""Template for a custom Prep PackageEndpoint"""
from __future__ import annotations

import zope.interface

import colrev.package_manager.interfaces
import colrev.package_manager.package_settings
import colrev.process.operation
from colrev.constants import Fields


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class CustomPrep:
    """Class for custom prep scripts"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    source_correction_hint = "check with the developer"
    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)

    def prepare(
        self,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        record: colrev.record.record.Record,
    ) -> colrev.record.record.Record:
        """Update record (metadata)"""

        if Fields.JOURNAL in record.data:
            if record.data[Fields.JOURNAL] == "MISQ":
                record.update_field(
                    key=Fields.JOURNAL, value="MIS Quarterly", source="custom_prep"
                )

        return record



#! /usr/bin/env python
"""Convenience functions to load JSON files"""
from __future__ import annotations

import json
import logging
import typing
from pathlib import Path

import colrev.loader.loader


# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes


class JSONLoader(colrev.loader.loader.Loader):
    """Loads json files"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        filename: Path,
        entrytype_setter: typing.Callable,
        field_mapper: typing.Callable,
        id_labeler: typing.Callable,
        unique_id_field: str = "",
        logger: logging.Logger = logging.getLogger(__name__),
    ):
        super().__init__(
            filename=filename,
            id_labeler=id_labeler,
            unique_id_field=unique_id_field,
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=logger,
        )

    @classmethod
    def get_nr_records(cls, filename: Path) -> int:
        """Get the number of records in the file"""
        with open(filename, encoding="utf-8-sig") as file:
            records_list = json.load(file)

        return len(records_list)

    def load_records_list(self) -> list:
        """Load json entries"""

        with open(self.filename, encoding="utf-8-sig") as file:
            records_list = json.load(file)

        return records_list



#! /usr/bin/env python
"""Default deduplication module for CoLRev"""
from __future__ import annotations

import shutil
from pathlib import Path

import bib_dedupe.cluster
import bib_dedupe.maybe_cases
import pandas as pd
import zope.interface
from bib_dedupe.bib_dedupe import block
from bib_dedupe.bib_dedupe import export_maybe
from bib_dedupe.bib_dedupe import import_maybe
from bib_dedupe.bib_dedupe import match
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import RecordState

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.DedupeInterface)
class Dedupe:
    """Default deduplication"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=True)

    def __init__(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
        settings: dict,
    ):
        self.settings = self.settings_class(**settings)
        self.dedupe_operation = dedupe_operation
        self.review_manager = dedupe_operation.review_manager

    def _move_maybe_file(self) -> None:
        # Note : temporary until we can pass a target path to bib_dedupe
        maybe_file = self.review_manager.path / Path(
            bib_dedupe.maybe_cases.MAYBE_CASES_FILEPATH
        )
        target_path = self.review_manager.paths.dedupe / Path(
            bib_dedupe.maybe_cases.MAYBE_CASES_FILEPATH
        )
        if not maybe_file.is_file():
            return
        shutil.move(str(maybe_file), str(target_path))

    def run_dedupe(self) -> None:
        """Run default dedupe"""

        records = self.review_manager.dataset.load_records_dict()
        records_df = pd.DataFrame.from_dict(records, orient="index")
        records_df = records_df[
            ~(
                records_df[Fields.STATUS].isin(
                    [
                        "md_imported",
                        "md_needs_manual_preparation",
                    ]
                )
            )
        ]
        verbosity_level = 0
        if self.review_manager.verbose_mode:
            verbosity_level = 1
        records_df.loc[
            records_df[Fields.STATUS].isin(
                RecordState.get_post_x_states(state=RecordState.md_processed)
            ),
            "search_set",
        ] = "old_search"

        records_df = self.dedupe_operation.get_records_for_dedupe(
            records_df=records_df, verbosity_level=verbosity_level
        )

        if 0 == records_df.shape[0]:
            return

        deduplication_pairs = block(records_df, verbosity_level=verbosity_level)
        matched_df = match(deduplication_pairs, verbosity_level=verbosity_level)
        matched_df = import_maybe(matched_df)

        if self.dedupe_operation.debug:
            return

        duplicate_id_sets = bib_dedupe.cluster.get_connected_components(matched_df)

        self.dedupe_operation.apply_merges(
            id_sets=duplicate_id_sets, complete_dedupe=True
        )

        self.review_manager.dataset.create_commit(
            msg="Dedupe: merge duplicate records",
        )

        export_maybe(records_df, matched_df=matched_df)
        self._move_maybe_file()



#! /usr/bin/env python
"""Load conversion of reference sections (bibliographies) in md-documents based on GROBID"""
from __future__ import annotations

import logging
import typing
from pathlib import Path

import requests

import colrev.env.grobid_service
import colrev.loader.bib
import colrev.loader.loader
from colrev.constants import Fields

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code
# pylint: disable=too-many-arguments


class MarkdownLoader(colrev.loader.loader.Loader):
    """Loads reference strings from text (md) files (based on GROBID)"""

    def __init__(
        self,
        *,
        filename: Path,
        entrytype_setter: typing.Callable,
        field_mapper: typing.Callable,
        id_labeler: typing.Callable,
        unique_id_field: str = "",
        logger: logging.Logger = logging.getLogger(__name__),
    ):

        super().__init__(
            filename=filename,
            id_labeler=id_labeler,
            unique_id_field=unique_id_field,
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=logger,
        )

    @classmethod
    def get_nr_records(cls, filename: Path) -> int:
        """Get the number of records in the file"""

        count = 0
        with open(filename, encoding="utf8") as file:
            for line in file:
                line = line.strip()
                if line and not line.startswith("#"):
                    count += 1
        return count

    def load_records_list(self) -> list:
        """Load records from the source"""

        self.logger.info("Running GROBID to parse structured reference data")

        grobid_service = colrev.env.grobid_service.GrobidService()

        grobid_service.check_grobid_availability()
        with open(self.filename, encoding="utf8") as file:
            references = [line.rstrip() for line in file if "#" not in line[:2]]

        data = ""
        ind = 0
        for ref in references:
            options = {}
            options["consolidateCitations"] = "0"
            options["citations"] = ref
            ret = requests.post(
                grobid_service.GROBID_URL + "/api/processCitation",
                data=options,
                headers={"Accept": "application/x-bibtex"},
                timeout=30,
            )
            ind += 1
            data = data + "\n" + ret.text.replace("{-1,", "{" + str(ind) + ",")

        records_dict = colrev.loader.load_utils.loads(
            load_string=data,
            implementation="bib",
            logger=self.logger,
        )

        for record in records_dict.values():
            if record.get(Fields.YEAR, "a") == record.get("date", "b"):
                del record["date"]

        return list(records_dict.values())



#! /usr/bin/env python
"""CoLRev prep_man operation: Prepare metadata manually."""
from __future__ import annotations

import csv
from pathlib import Path

import pandas as pd
from tqdm import tqdm

import colrev.env.language_service
import colrev.exceptions as colrev_exceptions
import colrev.process.operation
import colrev.record.record_prep
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.constants import RecordState


class PrepMan(colrev.process.operation.Operation):
    """Prepare records manually (metadata)"""

    type = OperationsType.prep_man

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.verbose = True
        self.lang_prep_csv_path = self.review_manager.paths.prep / Path(
            "missing_lang_recs_df.csv"
        )

    def _get_crosstab_df(self) -> pd.DataFrame:
        # pylint: disable=too-many-branches
        # pylint: disable=duplicate-code

        records = self.review_manager.dataset.load_records_dict()

        self.review_manager.logger.info("Calculate statistics")
        stats: dict = {Fields.ENTRYTYPE: {}}
        overall_types: dict = {Fields.ENTRYTYPE: {}}
        prep_man_hints, origins, crosstab = [], [], []
        for record_dict in records.values():
            if RecordState.md_imported != record_dict[Fields.STATUS]:
                if record_dict[Fields.ENTRYTYPE] in overall_types[Fields.ENTRYTYPE]:
                    overall_types[Fields.ENTRYTYPE][record_dict[Fields.ENTRYTYPE]] = (
                        overall_types[Fields.ENTRYTYPE][record_dict[Fields.ENTRYTYPE]]
                        + 1
                    )
                else:
                    overall_types[Fields.ENTRYTYPE][record_dict[Fields.ENTRYTYPE]] = 1

            if RecordState.md_needs_manual_preparation != record_dict[Fields.STATUS]:
                continue

            if record_dict[Fields.ENTRYTYPE] in stats[Fields.ENTRYTYPE]:
                stats[Fields.ENTRYTYPE][record_dict[Fields.ENTRYTYPE]] = (
                    stats[Fields.ENTRYTYPE][record_dict[Fields.ENTRYTYPE]] + 1
                )
            else:
                stats[Fields.ENTRYTYPE][record_dict[Fields.ENTRYTYPE]] = 1

            if Fields.MD_PROV in record_dict:
                record = colrev.record.record_prep.PrepRecord(record_dict)
                prov_d = record.data[Fields.MD_PROV]
                hints = []
                for key, value in prov_d.items():
                    if value["note"] != "":
                        hints.append(f'{key} - {value["note"]}')

                prep_man_hints.append([hint.lstrip() for hint in hints])
                for hint in hints:
                    if "change-score" in hint:
                        continue
                    # Note: if something causes the needs_manual_preparation
                    # it is caused by all colrev_origins
                    for orig in record_dict.get(Fields.ORIGIN, ["NA"]):
                        crosstab.append([orig[: orig.rfind("/")], hint.lstrip()])

            origins.append(
                [x[: x.rfind("/")] for x in record_dict.get(Fields.ORIGIN, ["NA"])]
            )

        print("Entry type statistics overall:")
        self.review_manager.p_printer.pprint(overall_types[Fields.ENTRYTYPE])

        print("Entry type statistics (needs_manual_preparation):")
        self.review_manager.p_printer.pprint(stats[Fields.ENTRYTYPE])

        return pd.DataFrame(crosstab, columns=[Fields.ORIGIN, "hint"])

    def _export_prep_man_langs(self, records: dict) -> None:
        language_service = colrev.env.language_service.LanguageService()

        missing_lang_recs = []
        self.review_manager.logger.info(
            "Calculate most likely languages for records without language field"
        )
        for record in tqdm(records.values()):
            if Fields.TITLE not in record:
                continue
            if Fields.LANGUAGE not in record:
                confidence_values = language_service.compute_language_confidence_values(
                    text=record[Fields.TITLE]
                )
                predicted_language, conf = confidence_values.pop(0)

                lang_rec = {
                    Fields.ID: record[Fields.ID],
                    Fields.TITLE: record[Fields.TITLE],
                    "most_likely_language": predicted_language,
                    "confidence": conf,
                }
                missing_lang_recs.append(lang_rec)

        missing_lang_recs_df = pd.DataFrame(missing_lang_recs)
        missing_lang_recs_df.to_csv(
            self.lang_prep_csv_path, index=False, quoting=csv.QUOTE_ALL
        )
        self.review_manager.logger.info(f"Exported table to {self.lang_prep_csv_path}")
        self.review_manager.logger.info(
            "Update the language column and rerun colrev prep-man -l"
        )

    def _import_prep_man_langs(self, records: dict) -> None:
        self.review_manager.logger.info(
            f"Import language fields from {self.lang_prep_csv_path}"
        )
        languages_df = pd.read_csv(self.lang_prep_csv_path)
        language_records = languages_df.to_dict("records")
        for language_record in language_records:
            if language_record["most_likely_language"] == "":
                continue
            if language_record[Fields.ID] not in records:
                # warn
                continue
            record_dict = records[language_record[Fields.ID]]
            record = colrev.record.record_prep.PrepRecord(record_dict)
            record.update_field(
                key=Fields.LANGUAGE,
                value=language_record["most_likely_language"],
                source="LanguageDetector/Manual",
                note="",
            )

            if "language of title not in [eng]" == record.data.get(
                Fields.PRESCREEN_EXCLUSION, ""
            ):
                record.remove_field(key=Fields.PRESCREEN_EXCLUSION)
                record.remove_field_provenance_note(
                    key=Fields.TITLE, note="language-not-found"
                )
                if (
                    record.data[Fields.STATUS]
                    == RecordState.md_needs_manual_preparation
                ):
                    # by resetting to md_imported,
                    # the prescreen-exclusion based on languages will be reapplied.
                    record.set_status(RecordState.md_imported)

        self.review_manager.dataset.save_records_dict(records)

    def prep_man_langs(self) -> None:
        """Add missing language fields based on spreadsheets"""

        self.lang_prep_csv_path.parent.mkdir(exist_ok=True, parents=True)

        records = self.review_manager.dataset.load_records_dict()

        if not self.lang_prep_csv_path.is_file():
            self._export_prep_man_langs(records)

        else:
            self._import_prep_man_langs(records)

    def prep_man_stats(self) -> None:
        """Print statistics on prep_man"""
        # pylint: disable=duplicate-code

        crosstab_df = self._get_crosstab_df()

        if crosstab_df.empty:
            print("No records to prepare manually.")
        else:
            # pylint: disable=duplicate-code
            tabulated = pd.pivot_table(
                crosstab_df[[Fields.ORIGIN, "hint"]],
                index=[Fields.ORIGIN],
                columns=["hint"],
                aggfunc=len,
                fill_value=0,
                margins=True,
            )
            # .sort_index(axis='columns')
            tabulated.sort_values(by=["All"], ascending=False, inplace=True)
            # Transpose because we tend to have more error categories than search files.
            tabulated = tabulated.transpose()
            print(tabulated)
            self.review_manager.logger.info(
                "Writing data to file: manual_preparation_statistics.csv"
            )
            tabulated.to_csv("manual_preparation_statistics.csv")

    def get_data(self) -> dict:
        """Get the data for prep-man"""
        # pylint: disable=duplicate-code

        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())
        nr_tasks = len(
            [
                x
                for x in record_header_list
                if RecordState.md_needs_manual_preparation == x[Fields.STATUS]
            ]
        )

        all_ids = [x[Fields.ID] for x in record_header_list]

        pad = min((max(len(x[Fields.ID]) for x in record_header_list) + 2), 35)

        items = self.review_manager.dataset.read_next_record(
            conditions=[{Fields.STATUS: RecordState.md_needs_manual_preparation}]
        )

        md_prep_man_data = {
            "nr_tasks": nr_tasks,
            "items": items,
            "all_ids": all_ids,
            "PAD": pad,
        }
        self.review_manager.logger.debug(
            self.review_manager.p_printer.pformat(md_prep_man_data)
        )
        return md_prep_man_data

    def set_data(self, *, record_dict: dict) -> None:
        """Set data in the prep_man operation"""

        record = colrev.record.record_prep.PrepRecord(record_dict)
        record.set_masterdata_complete(
            source="prep_man",
            masterdata_repository=self.review_manager.settings.is_curated_repo(),
        )
        record.set_masterdata_consistent()
        # record.set_fields_complete()
        record.set_status(RecordState.md_prepared)
        record_dict = record.get_data()

        self.review_manager.dataset.save_records_dict(
            {record_dict[Fields.ID]: record_dict}, partial=True
        )

    @colrev.process.operation.Operation.decorate()
    def main(self) -> None:
        """Manually prepare records (main entrypoint)"""

        if (
            self.review_manager.in_ci_environment()
            and not self.review_manager.in_test_environment()
        ):
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="colrev prep-man",
                detailed_trace="prep-man not available in ci environment",
            )

        records = self.review_manager.dataset.load_records_dict()

        package_manager = self.review_manager.get_package_manager()

        for (
            prep_man_package_endpoint
        ) in self.review_manager.settings.prep.prep_man_package_endpoints:
            prep_man_class = package_manager.get_package_endpoint_class(
                package_type=EndpointType.prep_man,
                package_identifier=prep_man_package_endpoint["endpoint"],
            )
            endpoint = prep_man_class(
                prep_man_operation=self, settings=prep_man_package_endpoint
            )
            records = endpoint.prepare_manual(records)  # type: ignore



#!/usr/bin/env python3
"""Scripts to print the CoLRev status (cli)."""
from __future__ import annotations

import difflib
import subprocess  # nosec
import typing

import inquirer
from rapidfuzz import fuzz

import colrev.record.record
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import RecordState

# pylint: disable=duplicate-code
keys = [
    Fields.AUTHOR,
    Fields.TITLE,
    Fields.JOURNAL,
    Fields.BOOKTITLE,
    Fields.YEAR,
    Fields.VOLUME,
    Fields.NUMBER,
    Fields.PAGES,
]


def print_string_diff(change: tuple) -> str:
    """Generates a string representation of the differences between two strings."""
    diff = difflib.Differ()
    letters = list(diff.compare(change[1], change[0]))
    for i, letter in enumerate(letters):
        if letter.startswith("  "):
            letters[i] = letters[i][-1]
        elif letter.startswith("+ "):
            letters[i] = f"{Colors.RED}" + letters[i][-1] + f"{Colors.END}"
        elif letter.startswith("- "):
            letters[i] = f"{Colors.GREEN}" + letters[i][-1] + f"{Colors.END}"
    res = "".join(letters).replace("\n", " ")
    return res


def print_diff(origin: dict, record_dict: dict) -> None:
    """Print the diff between two records"""

    for key in keys:

        if key not in origin and key not in record_dict:
            continue
        if key not in origin and key in record_dict:
            print(f"  diff {key}: {Colors.GREEN}{record_dict.get(key, '')}{Colors.END}")
            continue
        if key in origin and key not in record_dict:
            print(f"  diff {key}: {Colors.RED}{record_dict.get(key, '')}{Colors.END}")
            continue
        if origin[key] == record_dict[key]:
            continue

        similarity = fuzz.partial_ratio(origin[key], record_dict[key]) / 100
        if similarity < 0.5 or key in [
            Fields.VOLUME,
            Fields.NUMBER,
            Fields.YEAR,
        ]:
            line = f"{origin[key]} > {Colors.RED}{record_dict[key]}{Colors.END}"
        else:
            line = print_string_diff((origin[key], record_dict[key]))
        print(f"  diff {key} : {line}")


def print_diff_pair(record_pair: list) -> None:
    """Print the diff between two records"""

    for key in keys:
        prev_val = "_FIRST_VAL"
        for rec in record_pair:
            if (
                prev_val == rec.get(key, FieldValues.UNKNOWN)
                or prev_val == "_FIRST_VAL"
            ):
                line = f"{rec.get(key, '')}"
            else:
                similarity = 0.0
                if (
                    prev_val != FieldValues.UNKNOWN
                    and rec.get(key, "") != ""
                    and rec[key] is not None
                ):
                    similarity = fuzz.partial_ratio(prev_val, rec[key]) / 100
                    # Note : the fuzz.partial_ratio works better for partial substrings
                    # from difflib import SequenceMatcher
                    # similarity = SequenceMatcher(None, prev_val, rec[key]).ratio()
                if (
                    prev_val == FieldValues.UNKNOWN
                    and rec.get(key, FieldValues.UNKNOWN) != FieldValues.UNKNOWN
                ):
                    line = f"{Colors.GREEN}{rec.get(key, '')}{Colors.END}"
                elif (
                    rec.get(key, FieldValues.UNKNOWN) == FieldValues.UNKNOWN
                    and prev_val != FieldValues.UNKNOWN
                ):
                    line = f"{Colors.RED}[REMOVED]{Colors.END}"
                elif similarity < 0.5 or key in [
                    Fields.VOLUME,
                    Fields.NUMBER,
                    Fields.YEAR,
                ]:
                    line = f"{Colors.RED}{rec.get(key, '')}{Colors.END}"
                else:
                    line = print_string_diff((prev_val, rec.get(key, "")))
            print(f"{key} : {line}")
            prev_val = rec.get(key, FieldValues.UNKNOWN)
        print()


def _validate_dedupe(
    *,
    validate_operation: colrev.ops.validate.Validate,
    validation_details: dict,
    threshold: float,  # pylint: disable=unused-argument
) -> None:
    dedupe_operation = validate_operation.review_manager.get_dedupe_operation()

    for validation_item in validation_details:
        print(validation_item["change_score"])
        print_diff_pair(
            record_pair=[
                validation_item["prior_record_a"],
                validation_item["prior_record_b"],
            ],
        )

        user_selection = input("Validate [y,n,q for yes, no (undo), or quit]?")

        if user_selection == "n":
            dedupe_operation.unmerge_records(
                current_record_ids=validation_item["record"][Fields.ID]
            )

        if user_selection == "q":
            break
        if user_selection == "y":
            continue


def _validate_prep_prescreen_exclusions(
    *,
    validate_operation: colrev.ops.validate.Validate,
    validation_details: dict,
) -> None:

    prescreen_excluded_to_validate = validation_details
    if prescreen_excluded_to_validate:
        print("Prescreen excluded:")
        prescreen_errors = []

        choices = [
            (
                f"{e['ID']} : " f"{Colors.ORANGE}{e['title']}{Colors.END}",
                i,
            )
            for i, e in enumerate(prescreen_excluded_to_validate)
        ]
        questions = [
            inquirer.Checkbox(
                "selected_records",
                message="Select prescreen errors (using space)",
                choices=choices,
            ),
        ]
        answers = inquirer.prompt(questions)
        selected_indices = answers["selected_records"]
        for index in selected_indices:
            prescreen_errors.append(prescreen_excluded_to_validate[index])

        for error in prescreen_errors:
            colrev.record.record.Record(error).set_status(
                target_state=RecordState.md_needs_manual_preparation
            )
            validate_operation.review_manager.dataset.save_records_dict(
                {error[Fields.ID]: error},
                partial=True,
            )


# def _remove_from_feed(*, review_manager, origin) -> None:
#     feed_name, feed_id = origin.split("/")
#     feed_file = review_manager.paths.search / Path(feed_name)
#     records = colrev.loader.load_utils.load(
#         filename=feed_file,
#         logger=review_manager.logger,
#         unique_id_field="ID",
#     )
#     records = {k: v for k, v in records.items() if k != feed_id}
#     write_file(records_dict=records, filename=feed_file)


def _validate_prep(
    *,
    validate_operation: colrev.ops.validate.Validate,
    validation_details: dict,
    threshold: float,
) -> None:

    # Part 1 : origin validation
    origins_to_remove: typing.Dict[str, list] = {}
    displayed = False
    for validation_element in validation_details:
        if validation_element["change_score_max"] < threshold:
            continue
        displayed = True
        print("\n\n\n\n\n")
        # Escape sequence to clear terminal output for each new comparison
        # os.system("cls" if os.name == "nt" else "clear")

        record_dict = validation_element["record_dict"]
        print("Origins")
        for origin in validation_element["origins"]:
            print()
            print(f"{origin[Fields.ORIGIN][0]} : change {origin['change_score']}")
            colrev.record.record.Record(origin).print_citation_format()
            print_diff(
                origin=origin,
                record_dict=record_dict,
            )
            # break # we could continue in verbose mode!?

        print()
        print(f"Current record: {record_dict[Fields.ID]}")
        colrev.record.record.Record(record_dict).print_citation_format()
        print()

        questions = [
            inquirer.List(
                "user_selection",
                message="Validate?",
                choices=[
                    "yes",
                    "no (remove origin)",
                    "quit",
                ],
                default="y",
            ),
        ]
        answers = inquirer.prompt(questions)
        user_selection = answers["user_selection"]

        if user_selection.startswith("no (remove origin)"):
            # remove origin
            choices = [
                f"{origin[Fields.ORIGIN][0]}"
                for i, origin in enumerate(list(validation_element["origins"]))
            ]
            questions = [
                inquirer.Checkbox(
                    "selected_records",
                    message="Select origin to remove (using space)",
                    choices=choices,
                ),
            ]
            answers = inquirer.prompt(questions)
            for origin_to_remove in answers["selected_records"]:
                file, identifier = origin_to_remove.split("/")
                if file not in origins_to_remove:
                    origins_to_remove[file] = []
                origins_to_remove[file].append(identifier)

            record_dict[Fields.ORIGIN] = [
                origin[Fields.ORIGIN][0]
                for origin in list(validation_element["origins"])
                if origin[Fields.ORIGIN][0] not in answers["selected_records"]
            ]

            validate_operation.review_manager.dataset.save_records_dict(
                {record_dict[Fields.ID]: record_dict},
                partial=True,
            )

        if user_selection == "quit":
            break
        if user_selection == "yes":
            continue

    # if there is a key in origins_to_remove that does not start with md_, print
    # pylint: disable=use-a-generator
    if any([key for key in origins_to_remove if not key.startswith("md_")]):
        print("Please run colrev load again to (re)load all origin(s)")

    validate_operation.remove_md_origins(origins_to_remove)

    if not displayed:
        validate_operation.review_manager.logger.info(
            "No preparation changes above threshold"
        )


def _validate_properties(
    *,
    validate_operation: colrev.ops.validate.Validate,
    validation_details: dict,
) -> None:

    validate_operation.review_manager.logger.info(
        " Traceability of records".ljust(32, " ")
        + str(validation_details["record_traceability"])
    )
    validate_operation.review_manager.logger.info(
        " Consistency (based on hooks)".ljust(32, " ")
        + str(validation_details["consistency"])
    )
    validate_operation.review_manager.logger.info(
        " Completeness of iteration".ljust(32, " ")
        + str(validation_details["completeness"])
    )


def _validate_contributor_commits(
    *,
    validate_operation: colrev.ops.validate.Validate,
    validation_details: dict,
) -> None:
    validate_operation.review_manager.logger.info(
        "Showing commits in which the contributor was involved as the author or committer."
    )

    print()
    print("Commits to validate:")
    print()
    for item in validation_details:
        for _, item_values in item.items():
            print(item_values["msg"])
            print(f"  date      {item_values['date']}")
            print(
                f"  author    {item_values['author']} ({item_values['author_email']})"
            )
            print(
                f"  committer {item_values['committer']} ({item_values['committer_email']})"
            )
            print(f"  {Colors.ORANGE}{item_values['validate']}{Colors.END}")

    print()


def _validate_general(
    *,
    validate_operation: colrev.ops.validate.Validate,
    validation_details: dict,
) -> None:
    validate_operation.review_manager.logger.info("Start general validation")
    validate_operation.review_manager.logger.info(
        "Next, an interface will open and "
        "display the changes introduced in the selected commit."
    )
    validate_operation.review_manager.logger.info(
        "To undo minor changes, edit the corresponding files directly and run "
        f"{Colors.ORANGE}git add FILENAME && "
        f"git commit -m 'DESCRIPTION OF CHANGES UNDONE'{Colors.END}"
    )
    validate_operation.review_manager.logger.info(
        "To undo all changes introduced in a commit, run "
        f"{Colors.ORANGE}git revert COMMIT_ID{Colors.END}"
    )
    input("Enter to continue")
    if "commit_relative" in validation_details:
        subprocess.run(  # nosec
            ["gitk", f"--select-commit={validation_details['commit_relative']}"],
            check=False,
        )
    else:
        subprocess.run(["gitk"], check=False)  # nosec


def validate(
    *,
    validate_operation: colrev.ops.validate.Validate,
    validation_details: dict,
    threshold: float,
) -> None:
    """Validate details in the cli"""

    for key, details in validation_details.items():
        if key == "prep":
            _validate_prep(
                validate_operation=validate_operation,
                validation_details=details,
                threshold=threshold,
            )
        elif key == "prep_prescreen_exclusions":
            _validate_prep_prescreen_exclusions(
                validate_operation=validate_operation,
                validation_details=details,
            )
        elif key == "dedupe":
            _validate_dedupe(
                validate_operation=validate_operation,
                validation_details=details,
                threshold=threshold,
            )
        elif key == "properties":
            _validate_properties(
                validate_operation=validate_operation,
                validation_details=details,
            )
        elif key == "contributor_commits":
            _validate_contributor_commits(
                validate_operation=validate_operation,
                validation_details=details,
            )
        elif key == "general":
            _validate_general(
                validate_operation=validate_operation,
                validation_details=details,
            )
        else:
            print("Not yet implemented")
            print(validation_details)

    if validate_operation.review_manager.dataset.records_changed():
        validate_operation.review_manager.dataset.create_commit(msg="validate")



#! /usr/bin/env python
"""Checker for doi-not-matching-pattern."""
from __future__ import annotations

import re

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class DOIPatternChecker:
    """The DOIPatternChecker"""

    msg = DefectCodes.DOI_NOT_MATCHING_PATTERN
    # https://www.crossref.org/blog/dois-and-matching-regular-expressions/
    _DOI_REGEX = r"^10.\d{4,9}\/"

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the doi-not-matching-pattern checks"""

        if Fields.DOI not in record.data or record.ignored_defect(
            key=Fields.DOI, defect=self.msg
        ):
            return

        if not re.match(self._DOI_REGEX, record.data[Fields.DOI]):
            record.add_field_provenance_note(key=Fields.DOI, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.DOI, note=self.msg)


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(DOIPatternChecker(quality_model))
